from typing import AsyncIterator, Iterator, Optional

from nuclia_models.agent.interaction import AragAnswer

from nuclia.data import get_auth
from nuclia.decorators import agent
from nuclia.lib.agent import AgentClient, AsyncAgentClient
from nuclia.sdk.agent_cli import NucliaAgentCLI
from nuclia.sdk.agent_sessions import AsyncNucliaAgentSessions, NucliaAgentSessions
from nuclia.sdk.auth import NucliaAuth


class NucliaAgent:
    session: NucliaAgentSessions
    cli: NucliaAgentCLI

    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    def __init__(self) -> None:
        self.session = NucliaAgentSessions()
        self.cli = NucliaAgentCLI()

    @agent
    def interact(
        self,
        question: str,
        session_uuid: str = "ephemeral",
        headers: Optional[dict[str, str]] = None,
        **kwargs,
    ) -> Iterator[AragAnswer]:
        ac: AgentClient = kwargs["ac"]
        for response in ac.interact(session_uuid, question, headers=headers):
            yield response


class AsyncNucliaAgent:
    session: "AsyncNucliaAgentSessions"

    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    def __init__(self) -> None:
        self.session = AsyncNucliaAgentSessions()

    @agent
    async def interact(
        self,
        question: str,
        session_uuid: str = "ephemeral",
        headers: Optional[dict[str, str]] = None,
        **kwargs,
    ) -> AsyncIterator[AragAnswer]:
        ac: AsyncAgentClient = kwargs["ac"]
        async for response in ac.interact(session_uuid, question, headers=headers):
            yield response



from typing import Type, Union

import pytest
from nucliadb_models.graph.requests import (
    AnyNode,
    GraphSearchRequest,
    NodeMatchKindName,
)
from nucliadb_models.search import (
    AskRequest,
    CatalogRequest,
    CitationsType,
    CustomPrompt,
    FindRequest,
    Reasoning,
    SearchRequest,
)
from pydantic import ValidationError

from nuclia.sdk.search import AsyncNucliaSearch, NucliaSearch
from nuclia.tests.utils import maybe_await


@pytest.mark.parametrize(
    "search_klass",
    [NucliaSearch, AsyncNucliaSearch],
)
@pytest.mark.parametrize(
    "query",
    [
        "Who is hedy Lamarr?",
        {"query": "Who is hedy Lamarr?"},
        SearchRequest(query="Who is hedy Lamarr?"),
    ],
)
async def test_search(
    testing_config,
    search_klass: Union[Type[NucliaSearch], Type[AsyncNucliaSearch]],
    query,
):
    search = search_klass()
    maybe_results = search.search(query=query)
    results = await maybe_await(maybe_results)

    assert len(results.resources.keys()) == 2
    titles = [r.title for r in results.resources.values()]
    assert "Lamarr Lesson plan.pdf" in titles


@pytest.mark.parametrize(
    "search_klass",
    [NucliaSearch, AsyncNucliaSearch],
)
async def test_search_with_filters(
    testing_config,
    search_klass: Union[Type[NucliaSearch], Type[AsyncNucliaSearch]],
):
    search = search_klass()
    maybe_results = search.search(
        query="Who is hedy Lamarr?", filters=["/icon/application/pdf"]
    )
    results = await maybe_await(maybe_results)
    assert len(results.resources.keys()) == 1


@pytest.mark.parametrize(
    "search_klass",
    [NucliaSearch, AsyncNucliaSearch],
)
@pytest.mark.parametrize(
    "query",
    [
        "",
        {"query": ""},
        CatalogRequest(query=""),
    ],
)
async def test_catalog(
    testing_config,
    search_klass: Union[Type[NucliaSearch], Type[AsyncNucliaSearch]],
    query,
):
    search = search_klass()
    maybe_results = search.catalog(query=query)
    results = await maybe_await(maybe_results)
    assert len(results.resources.keys()) >= 2


@pytest.mark.parametrize(
    "search_klass",
    [NucliaSearch, AsyncNucliaSearch],
)
@pytest.mark.parametrize(
    "query",
    [
        "Who is hedy Lamarr?",
        {"query": "Who is hedy Lamarr?"},
        FindRequest(query="Who is Hedy Lamarr?"),
    ],
)
async def test_find(
    testing_config,
    search_klass: Union[Type[NucliaSearch], Type[AsyncNucliaSearch]],
    query,
):
    search = search_klass()
    maybe_results = search.find(query=query)
    results = await maybe_await(maybe_results)
    assert len(results.resources.keys()) == 2
    titles = [r.title for r in results.resources.values()]
    assert "Lamarr Lesson plan.pdf" in titles


@pytest.mark.parametrize(
    "search_klass",
    [NucliaSearch, AsyncNucliaSearch],
)
@pytest.mark.parametrize(
    "query",
    [
        "Who is hedy Lamarr?",
        {"query": "Who is hedy Lamarr?"},
        AskRequest(query="Who is Hedy Lamarr?"),
    ],
)
async def test_ask(
    testing_config,
    search_klass: Union[Type[NucliaSearch], Type[AsyncNucliaSearch]],
    query,
):
    search = search_klass()

    maybe_results = search.ask(query=query)
    results = await maybe_await(maybe_results)
    answer = results.answer.decode()
    assert "Lamarr" in answer


async def test_ask_with_custom_prompt(testing_config):
    search = NucliaSearch()
    async_search = AsyncNucliaSearch()

    ask = AskRequest(
        query="Are Pepito Palotes and Hedy Lamarr friends?",
        prompt=CustomPrompt(
            system="Answer the question. If you don't know the anwser, write 'I don't know' and a list of bullet points with the reasons why you don't know the answer.",
            user="Based on this context {context}, answer the question {question}",
        ),
        generative_model="chatgpt-azure-4o",
    )
    results = search.ask(query=ask)
    async_results = await async_search.ask(query=ask)
    assert results.find_result == async_results.find_result

    answer = results.answer.decode()
    assert "I don't know" in answer, print(answer)

    answer = async_results.answer.decode()
    assert "I don't know" in answer, print(answer)


def test_ask_with_markdown_answer(testing_config):
    search = NucliaSearch()
    ask = AskRequest(
        query="Who is Hedy Lamarr and what did she do?",
        prompt=CustomPrompt(
            system="Answer the question and use lists as much as possible.",
            user="Based on this context {context}, answer the question {question}",
        ),
        generative_model="gemini-2.5-pro",
        prefer_markdown=True,
    )
    results = search.ask(query=ask)
    answer = results.answer.decode()
    markdown_keywords = ["**", "#", "1.", "*"]
    assert any([keyword in answer.lower() for keyword in markdown_keywords]), answer


SCHEMA = {
    "name": "ClassificationReverse",
    "description": "Correctly extracted with all the required parameters with correct types",
    "parameters": {
        "$defs": {
            "Options": {
                "enum": ["SPORTS", "TECHNOLOGY"],
                "title": "Options",
                "type": "string",
            }
        },
        "properties": {
            "title": {"default": "label", "title": "Title", "type": "string"},
            "description": {
                "default": "Define labels to classify the subject of the document",
                "title": "Description",
                "type": "string",
            },
            "document_type": {
                "description": "Type of document, SPORT example: elections, Illa, POLITICAL example: football, TECHNOLOGY example: computer",
                "items": {"$ref": "#/$defs/Options"},
                "title": "Document Type",
                "type": "array",
            },
        },
        "required": ["document_type"],
        "type": "object",
    },
}


def test_ask_json(testing_config):
    search = NucliaSearch()
    results = search.ask_json(
        query="Who is hedy Lamarr?", filters=["/icon/application/pdf"], schema=SCHEMA
    )

    assert "TECHNOLOGY" in results.object["document_type"]


@pytest.mark.asyncio
async def test_ask_json_async(testing_config):
    search = AsyncNucliaSearch()
    results = await search.ask_json(
        query="Who is hedy Lamarr?",
        filters=["/icon/application/pdf"],
        schema=SCHEMA,
        show_consumption=True,
    )

    assert "TECHNOLOGY" in results.object["document_type"]
    assert results.consumption is not None


@pytest.mark.parametrize(
    "search_klass",
    [NucliaSearch, AsyncNucliaSearch],
)
@pytest.mark.parametrize(
    "query",
    [
        {
            "query": {
                "prop": "node",
                "value": "Hedy",
                "match": "fuzzy",
            }
        },
        GraphSearchRequest(query=AnyNode(value="Hedy", match=NodeMatchKindName.FUZZY)),
    ],
)
async def test_graph(
    testing_config,
    search_klass: Union[Type[NucliaSearch], Type[AsyncNucliaSearch]],
    query,
):
    search = search_klass()
    maybe_results = search.graph(query=query)
    results = await maybe_await(maybe_results)
    assert len(results.paths) >= 22


async def test_search_query_type_check(
    testing_config,
) -> None:
    query = object()  # invalid on every call

    search = NucliaSearch()
    with pytest.raises(TypeError):
        search.search(query=query)
    with pytest.raises(TypeError):
        search.find(query=query)
    with pytest.raises(TypeError):
        search.catalog(query=query)
    with pytest.raises(TypeError):
        search.ask(query=query)
    with pytest.raises(TypeError):
        search.ask_json(schema={}, query=query)
    with pytest.raises(TypeError):
        search.graph(query=query)

    async_search = AsyncNucliaSearch()
    with pytest.raises(TypeError):
        await async_search.search(query=query)
    with pytest.raises(TypeError):
        await async_search.find(query=query)
    with pytest.raises(TypeError):
        await async_search.catalog(query=query)
    with pytest.raises(TypeError):
        await async_search.ask(query=query)
    with pytest.raises(TypeError):
        async for _ in async_search.ask_stream(query=query):
            pass
    with pytest.raises(TypeError):
        await async_search.ask_json(schema={}, query=query)
    with pytest.raises(TypeError):
        await async_search.graph(query=query)


async def test_search_query_validation_errors(
    testing_config,
) -> None:
    query = {"query": object()}  # invalid on every call

    search = NucliaSearch()
    with pytest.raises(ValidationError):
        search.search(query=query)
    with pytest.raises(ValidationError):
        search.find(query=query)
    with pytest.raises(ValidationError):
        search.catalog(query=query)
    with pytest.raises(ValidationError):
        search.ask(query=query)
    with pytest.raises(ValidationError):
        search.ask_json(schema={}, query=query)
    with pytest.raises(ValidationError):
        search.graph(query=query)

    async_search = AsyncNucliaSearch()
    with pytest.raises(ValidationError):
        await async_search.search(query=query)
    with pytest.raises(ValidationError):
        await async_search.find(query=query)
    with pytest.raises(ValidationError):
        await async_search.catalog(query=query)
    with pytest.raises(ValidationError):
        await async_search.ask(query=query)
    with pytest.raises(ValidationError):
        async for _ in async_search.ask_stream(query=query):
            pass
    with pytest.raises(ValidationError):
        await async_search.ask_json(schema={}, query=query)
    with pytest.raises(ValidationError):
        await async_search.graph(query=query)


@pytest.mark.parametrize(
    "search_klass",
    [NucliaSearch, AsyncNucliaSearch],
)
@pytest.mark.parametrize(
    "query",
    [
        AskRequest(
            query=(
                "If the Navy had adopted Lamarr’s invention during World War II, how might this have changed the effectiveness of"
                " Allied naval communications and torpedo guidance? Support your reasoning with historical and technological context."
            ),
            reasoning=Reasoning(display=True, effort="high", budget_tokens=1024),
            generative_model="chatgpt-azure-o3-mini",
            max_tokens=5000,
        ),
    ],
)
async def test_ask_with_reasoning(
    testing_config,
    search_klass: Union[Type[NucliaSearch], Type[AsyncNucliaSearch]],
    query,
):
    search = search_klass()

    maybe_results = search.ask(query=query)
    results = await maybe_await(maybe_results)
    answer = results.answer.decode()
    assert "Lamarr" in answer
    # Reasoning is not very consistent since the model decides when to use it
    # assert results.reasoning is not None


@pytest.fixture(scope="function")
def ask_example_with_citations() -> AskRequest:
    return AskRequest(
        query="Who is hedy Lamarr?",
        citations=CitationsType.LLM_FOOTNOTES,
        generative_model="chatgpt-azure-4o-mini",
        max_tokens=1500,
    )


def test_ask_citation_footnote_to_context(testing_config, ask_example_with_citations):
    search = NucliaSearch()
    results = search.ask(query=ask_example_with_citations)

    assert results.citation_footnote_to_context, results.citation_footnote_to_context


@pytest.mark.asyncio
async def test_async_ask_citation_footnote_to_context(
    testing_config, ask_example_with_citations
):
    search = AsyncNucliaSearch()
    results = await search.ask(query=ask_example_with_citations)

    assert results.citation_footnote_to_context, results.citation_footnote_to_context



from typing import Dict

from nuclia_models.config.proto import ExtractConfig

from nuclia.data import get_async_auth, get_auth
from nuclia.decorators import kb, pretty
from nuclia.lib.kb import AsyncNucliaDBClient, NucliaDBClient
from nuclia.sdk.auth import AsyncNucliaAuth, NucliaAuth


class NucliaExtractStrategy:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @kb
    @pretty
    def list(self, *args, **kwargs) -> Dict[str, ExtractConfig]:
        """
        List extract strategies
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.list_extract_strategies()
        return response.json()

    @kb
    def add(
        self,
        *args,
        config: ExtractConfig,
        **kwargs,
    ) -> str:
        """
        Add extract strategy

        :param config: strategy configuration
        """
        if isinstance(config, dict):
            config = ExtractConfig.model_validate(config)

        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.add_extract_strategy(config=config)
        return response.json()

    @kb
    def delete(self, *args, id: str, **kwargs):
        """
        Delete extract strategy

        :param id: ID of the strategy to delete
        """
        ndb: NucliaDBClient = kwargs["ndb"]
        ndb.delete_extract_strategy(strategy_id=id)


class AsyncNucliaExtractStrategy:
    @property
    def _auth(self) -> AsyncNucliaAuth:
        auth = get_async_auth()
        return auth

    @kb
    @pretty
    async def list(self, *args, **kwargs) -> Dict[str, ExtractConfig]:
        """
        List extract strategies
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.list_extract_strategies()
        return response.json()

    @kb
    async def add(
        self,
        *args,
        config: ExtractConfig,
        **kwargs,
    ) -> str:
        """
        Add extract strategy

        :param config: strategy configuration
        """
        if isinstance(config, dict):
            config = ExtractConfig.model_validate(config)

        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.add_extract_strategy(config=config)
        return response.json()

    @kb
    async def delete(self, *args, id: str, **kwargs):
        """
        Delete extract strategy

        :param id: ID of the strategy to delete
        """
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        await ndb.delete_extract_strategy(strategy_id=id)



from enum import Enum
from typing import TYPE_CHECKING, Optional

from nuclia_models.agent.interaction import AnswerOperation, AragAnswer
from nuclia_models.agent.memory import Answer, Context, Step
from rich import box
from rich.console import Console, Group
from rich.live import Live
from rich.markdown import Markdown
from rich.panel import Panel
from rich.spinner import Spinner
from rich.table import Table

from nuclia.data import get_auth
from nuclia.decorators import agent
from nuclia.lib.agent import AgentClient
from nuclia.sdk.agent_sessions import NucliaAgentSessions

if TYPE_CHECKING:
    from nuclia.lib.agent import AgentClient
    from nuclia.sdk.agent_sessions import NucliaAgentSessions
    from nuclia.sdk.auth import NucliaAuth


class AgentCommand(Enum):
    EXIT = "exit"
    HELP = "help"
    NEW_SESSION = "new_session"
    LIST_SESSIONS = "list_sessions"
    CHANGE_SESSION = "change_session"
    CLEAR = "clear"


HELP = {
    AgentCommand.EXIT: "Exit the interactive CLI.",
    AgentCommand.HELP: "Show this help message.",
    AgentCommand.NEW_SESSION: "Create a new persistent session.",
    AgentCommand.LIST_SESSIONS: "List all created sessions.",
    AgentCommand.CHANGE_SESSION: "Change to a different session.",
    AgentCommand.CLEAR: "Clear the screen.",
}

MAX_RESULT_DISPLAY_LENGTH = 500
MAX_CONTEXT_DISPLAY_LENGTH = 300


class NucliaAgentCLI:
    # underscore so that it does not show up in the CLI commands
    _session: "NucliaAgentSessions"
    _console: Optional[Console] = None

    @property
    def _auth(self) -> "NucliaAuth":
        auth = get_auth()
        return auth

    def __init__(self, console: Optional[Console] = None) -> None:
        self._session = NucliaAgentSessions()
        self._console = console

    def _build_step_display(self, step: Step) -> Table:
        """Build a table for displaying a processing step."""
        step_info = Table.grid(padding=(0, 2))
        step_info.add_column(style="cyan", justify="right")
        step_info.add_column(style="white")

        step_info.add_row("Module:", f"[bold]{step.module}[/]")
        if step.value:
            step_info.add_row(
                "Result:",
                Markdown(
                    step.value[:MAX_RESULT_DISPLAY_LENGTH] + "..."
                    if len(step.value) > MAX_RESULT_DISPLAY_LENGTH
                    else step.value
                ),
            )
        if step.reason:
            step_info.add_row(
                "Reason:",
                Markdown(
                    step.reason[:MAX_RESULT_DISPLAY_LENGTH] + "..."
                    if len(step.reason) > MAX_RESULT_DISPLAY_LENGTH
                    else step.reason
                ),
            )
        step_info.add_row("Time:", f"{step.timeit:.2f}s")
        if step.input_nuclia_tokens or step.output_nuclia_tokens:
            step_info.add_row(
                "Tokens:",
                f"in: {step.input_nuclia_tokens:.0f}, out: {step.output_nuclia_tokens:.0f}",
            )
        return step_info

    def _build_context_display(self, ctx: Context) -> Table:
        """Build a table for displaying retrieved context."""
        context_table = Table(
            show_header=True,
            header_style="bold cyan",
            expand=True,
            show_edge=False,
            padding=(0, 1),
            box=box.SIMPLE,
        )
        context_table.add_column("Source", style="cyan", no_wrap=True, ratio=1)
        context_table.add_column("Content", style="white", ratio=7)

        for i, chunk in enumerate(ctx.chunks[:5]):  # Show first 5 chunks
            content = (
                chunk.text[:MAX_CONTEXT_DISPLAY_LENGTH] + "..."
                if len(chunk.text) > MAX_CONTEXT_DISPLAY_LENGTH
                else chunk.text
            )
            context_table.add_row(chunk.title or f"Chunk {i + 1}", content)

        # Add summary row if available
        if ctx.summary:
            context_table.add_row(
                "[bold]Summary[/]",
                ctx.summary,
            )
        return context_table

    def _build_error_display(self, exception) -> Markdown:
        """Build a formatted error message display."""
        error_message = f"**Error Details:**\n\n{exception.detail}"
        return Markdown(error_message)

    def _build_feedback_display(self, feedback) -> Table:
        """Build a table for displaying agent feedback request."""
        feedback_table = Table.grid(padding=(0, 2))
        feedback_table.add_column(style="cyan", justify="right")
        feedback_table.add_column(style="white")

        feedback_table.add_row("Module:", f"[bold]{feedback.module}[/]")
        feedback_table.add_row("Question:", feedback.question)
        if feedback.data:
            import json

            feedback_table.add_row("Data:", json.dumps(feedback.data, indent=2))

        return feedback_table

    def _build_possible_answer_display(self, possible_answer: Answer) -> Table:
        """Build a table for displaying a possible answer."""
        answer_table = Table.grid(padding=(0, 2))
        answer_table.add_column(style="cyan", justify="right")
        answer_table.add_column(style="white")

        if possible_answer.answer:
            answer_table.add_row("Text:", Markdown(possible_answer.answer))

        return answer_table

    def _print_panel(
        self, console: Console, content, title: str, border_style: str = "green"
    ):
        """Helper to print a panel with consistent agent title formatting."""
        console.print(
            Panel(
                content,
                title=title,
                border_style=border_style,
            )
        )

    def _print_welcome(self, console: Console, agent_title: str):
        """Print the welcome panel."""
        self._print_panel(
            console,
            "[bold]Nuclia Agent CLI[/]\n\n"
            "Type your questions to interact with the agent.\n"
            "Type [cyan]/help[/] for available commands.",
            f"[bold green]{agent_title}[/]",
            "green",
        )

    def _handle_new_session(
        self, console: Console, ac: "AgentClient", agent_title: str
    ) -> Optional[str]:
        """Handle creating a new session."""
        session_name = console.input("[bold cyan]Session name:[/] ").strip()
        if session_name:
            try:
                session_uuid = self._session.new(name=session_name, ac=ac)
                self._print_panel(
                    console,
                    f"[green]✓ Created session:[/] {session_name}\n[dim]UUID: {session_uuid}[/]",
                    f"[bold green]{agent_title}[/] · Session Created",
                    "green",
                )
                return session_uuid
            except Exception as e:
                self._print_panel(
                    console,
                    f"[red]✗ Failed to create session:[/] {str(e)}",
                    f"[bold green]{agent_title}[/] · Error",
                    "red",
                )
        return None

    def _handle_list_sessions(
        self, console: Console, ac: "AgentClient", agent_title: str
    ):
        """Handle listing all sessions."""
        try:
            sessions = self._session.list(ac=ac)
            if not sessions.resources:
                self._print_panel(
                    console,
                    "[yellow]No sessions found.[/]",
                    f"[bold green]{agent_title}[/] · Sessions",
                    "green",
                )
                return

            sessions_table = Table(
                show_header=True,
                header_style="bold magenta",
                expand=True,
                box=box.SIMPLE,
            )
            sessions_table.add_column("Title", style="cyan")
            sessions_table.add_column("UUID", style="dim")
            sessions_table.add_column("Created", style="green")

            for session in sessions.resources:
                sessions_table.add_row(
                    session.title or "Untitled",
                    session.id,
                    str(
                        session.created.strftime("%Y-%m-%d %H:%M")
                        if session.created
                        else "N/A"
                    ),
                )

            self._print_panel(
                console,
                sessions_table,
                f"[bold green]{agent_title}[/] · Available Sessions",
                "green",
            )
        except Exception as e:
            self._print_panel(
                console,
                f"[red]✗ Failed to list sessions:[/] {str(e)}",
                f"[bold green]{agent_title}[/] · Error",
                "red",
            )

    def _handle_change_session(
        self, console: Console, ac: "AgentClient", agent_title: str
    ) -> Optional[str]:
        """Handle changing to a different session."""
        new_session_uuid = console.input(
            "[bold cyan]Session UUID (or 'ephemeral'):[/] "
        ).strip()
        if new_session_uuid == "ephemeral":
            self._print_panel(
                console,
                "[green]✓ Switched to ephemeral session[/]",
                f"[bold green]{agent_title}[/] · Session Changed",
                "green",
            )
            return "ephemeral"
        elif new_session_uuid:
            try:
                self._session.get(session_uuid=new_session_uuid, ac=ac)
                self._print_panel(
                    console,
                    f"[green]✓ Switched to session:[/] {new_session_uuid}",
                    f"[bold green]{agent_title}[/] · Session Changed",
                    "green",
                )
                return new_session_uuid
            except Exception as e:
                self._print_panel(
                    console,
                    f"[red]✗ Failed to switch session:[/] {str(e)}",
                    f"[bold green]{agent_title}[/] · Error",
                    "red",
                )
        return None

    def _handle_command(
        self, command: str, console: Console, ac: "AgentClient", agent_title: str
    ) -> tuple[bool, Optional[str]]:
        """
        Handle a CLI command.

        Returns:
            Tuple of (should_exit, new_session_uuid)
        """
        if command == AgentCommand.EXIT.value:
            console.print("\n[yellow]Goodbye![/]")
            return True, None

        if command == AgentCommand.HELP.value:
            help_text = "[bold]Available Commands:[/]\n\n"
            for cmd, desc in HELP.items():
                help_text += f"  [cyan]/{cmd.value}[/] - {desc}\n"
            self._print_panel(
                console,
                help_text,
                f"[bold green]{agent_title}[/] · Help",
                "green",
            )
            return False, None

        if command == AgentCommand.NEW_SESSION.value:
            new_uuid = self._handle_new_session(console, ac, agent_title)
            return False, new_uuid

        if command == AgentCommand.LIST_SESSIONS.value:
            self._handle_list_sessions(console, ac, agent_title)
            return False, None

        if command == AgentCommand.CHANGE_SESSION.value:
            new_uuid = self._handle_change_session(console, ac, agent_title)
            return False, new_uuid

        if command == AgentCommand.CLEAR.value:
            console.clear()
            self._print_welcome(console, agent_title)
            console.print()
            return False, None

        # Unknown command
        self._print_panel(
            console,
            f"[red]Unknown command:[/] /{command}\n\nType [cyan]/help[/] for available commands.",
            f"[bold green]{agent_title}[/] · Error",
            "red",
        )
        return False, None

    def _update_live_display(
        self,
        live: Live,
        response_elements: list,
        agent_title: str,
        spinner: Spinner,
        show_spinner: bool = True,
    ):
        """Update the live display with current response elements."""
        if show_spinner:
            content = Group(*response_elements, spinner)
        else:
            content = Group(*response_elements)

        live.update(
            Panel(
                content,
                title=f"[bold green]{agent_title}[/]",
                border_style="green",
            )
        )

    def _handle_agent_feedback(
        self,
        response: AragAnswer,
        response_elements: list,
        live: Live,
        console: Console,
        agent_title: str,
        spinner: Spinner,
    ) -> str:
        """
        Handle agent request for user feedback.

        Returns the user's response string.
        """
        feedback_display = self._build_feedback_display(response.feedback)
        element = Panel(
            feedback_display,
            title="[bold magenta]🤖 Agent Request[/]",
            border_style="magenta",
        )
        response_elements.append(element)

        # Update display to show request
        live.update(
            Panel(
                Group(*response_elements),
                title=f"[bold magenta]{agent_title}[/] · Requesting Input",
                border_style="magenta",
            )
        )

        # Get user input
        live.stop()
        user_response = console.input("[bold magenta]Your response:[/] ").strip()
        live.start()

        # Add user response to display
        user_element = Panel(
            user_response,
            title="[bold blue]👤 Your Response[/]",
            border_style="blue",
        )
        response_elements.append(user_element)
        self._update_live_display(live, response_elements, agent_title, spinner)

        return user_response

    def _process_response_operation(
        self, response: AragAnswer, spinner: Spinner
    ) -> tuple[Optional[Panel], bool]:
        """
        Process a single response operation and return the display element and whether to stop.

        Returns:
            Tuple of (element to display or None, should_stop)
        """
        if response.operation == AnswerOperation.START:
            spinner.update(text="[dim]Starting[/]")
            return None, False

        if response.operation == AnswerOperation.DONE:
            spinner.update(text="[dim]Done[/]")
            return None, True

        if response.operation == AnswerOperation.ERROR or response.exception:
            if response.exception:
                error_display = self._build_error_display(response.exception)
                element = Panel(
                    error_display,
                    title="[bold red]❌ Error[/]",
                    border_style="red",
                )
                return element, True
            return None, True

        if response.step:
            step_info = self._build_step_display(response.step)
            element = Panel(
                step_info,
                title=f"[yellow]⚙️ {response.step.title}[/]",
                border_style="yellow",
            )
            spinner.update(text="[dim]Running[/]")
            return element, False

        if response.context:
            context_table = self._build_context_display(response.context)
            element = Panel(
                context_table,
                title="[bold cyan]📚 Context Retrieved[/]",
                border_style="cyan",
            )
            spinner.update(text="[dim]Running[/]")
            return element, False

        if response.possible_answer:
            possible_answer_display = self._build_possible_answer_display(
                response.possible_answer
            )
            element = Panel(
                possible_answer_display,
                title="[bold yellow]💭 Possible Answer[/]",
                border_style="yellow",
            )
            spinner.update(text="[dim]Generating[/]")
            return element, False

        if response.generated_text:
            element = Panel(
                Markdown(response.generated_text),
                title="[bold white]📝 Generated Text[/]",
                border_style="white",
            )
            spinner.update(text="[dim]Generating[/]")
            return element, False

        if response.answer:
            element = Panel(
                Markdown(response.answer),
                title="[bold green]✅ Answer[/]",
                border_style="green",
            )
            return element, True

        return None, False

    def _process_agent_question(
        self,
        ac: "AgentClient",
        session_uuid: str,
        question: str,
        console: Console,
        agent_title: str,
    ):
        """Process a user question through the agent."""
        response_elements: list = []
        spinner = Spinner("dots", text="[dim]Thinking...[/]", style="cyan")

        with Live(
            Panel(
                spinner,
                title=f"[bold green]{agent_title}[/]",
                border_style="green",
            ),
            console=console,
            refresh_per_second=10,
        ) as live:
            generator = ac.interact(session_uuid, question)
            user_feedback = None

            try:
                while True:
                    # Get next response (send user feedback if available)
                    if user_feedback is not None:
                        response = generator.send(user_feedback)
                        user_feedback = None
                    else:
                        response = next(generator)

                    # Handle agent feedback request
                    if (
                        response.operation == AnswerOperation.AGENT_REQUEST
                        and response.feedback
                    ):
                        user_feedback = self._handle_agent_feedback(
                            response,
                            response_elements,
                            live,
                            console,
                            agent_title,
                            spinner,
                        )
                        continue

                    # Process the response
                    element, should_stop = self._process_response_operation(
                        response, spinner
                    )

                    if element:
                        response_elements.append(element)
                        # Special handling for final answer or error
                        if should_stop:
                            self._update_live_display(
                                live,
                                response_elements,
                                agent_title,
                                spinner,
                                show_spinner=False,
                            )
                        else:
                            self._update_live_display(
                                live, response_elements, agent_title, spinner
                            )

                    if should_stop:
                        break

            except StopIteration:
                pass

        console.print()  # Add spacing after response

    def _get_session_display_name(
        self,
        session_uuid: str,
        session_titles: dict[str, Optional[str]],
        ac: "AgentClient",
    ) -> str:
        """Get display name for a session, fetching it if needed."""
        if session_uuid not in session_titles and session_uuid != "ephemeral":
            try:
                session_info = self._session.get(session_uuid=session_uuid, ac=ac)
                session_titles[session_uuid] = session_info.title
            except Exception:
                session_titles[session_uuid] = None

        return session_titles.get(session_uuid) or session_uuid[:8]

    @agent
    def interact(self, **kwargs):
        """Main interactive CLI loop."""
        ac: AgentClient = kwargs["ac"]
        agent_config = self._auth._config.get_agent(ac.agent_id)
        agent_title: str = (
            agent_config.title if agent_config and agent_config.title else "Agent"
        )
        session_uuid = "ephemeral"
        console = self._console or Console()
        session_titles: dict[str, Optional[str]] = {"ephemeral": "Ephemeral Session"}

        console.clear()
        self._print_welcome(console, agent_title)
        console.print()

        try:
            while True:
                # Display prompt with session info
                session_display = self._get_session_display_name(
                    session_uuid, session_titles, ac
                )
                prompt_text = f"[bold blue]You[/] [dim]({session_display}):[/] "
                user_input = console.input(prompt_text).strip()

                if not user_input:
                    continue

                # Handle commands
                if user_input.startswith("/"):
                    command = user_input[1:].lower()
                    should_exit, new_session_uuid = self._handle_command(
                        command, console, ac, agent_title
                    )
                    if should_exit:
                        break
                    if new_session_uuid:
                        session_uuid = new_session_uuid
                    continue

                # Process agent question
                try:
                    self._process_agent_question(
                        ac, session_uuid, user_input, console, agent_title
                    )
                except KeyboardInterrupt:
                    console.print(
                        Panel(
                            "[yellow]Interrupted[/]",
                            title=f"[bold green]{agent_title}[/]",
                            border_style="yellow",
                        )
                    )
                except Exception as e:
                    console.print(
                        Panel(
                            f"[red]✗ Error:[/] {str(e)}",
                            title=f"[bold green]{agent_title}[/]",
                            border_style="red",
                        )
                    )

        except KeyboardInterrupt:
            console.print("\n\n[yellow]Interrupted. Goodbye![/]")
        except Exception as e:
            console.print(f"\n[red]✗ Fatal Error:[/] {str(e)}")



import pytest
from nuclia_models.predict.generative_responses import (
    ConsumptionGenerative,
    FootnoteCitationsGenerativeResponse,
    TextGenerativeResponse,
)
from nuclia_models.predict.remi import RemiRequest

from nuclia.lib.nua_responses import (
    ChatModel,
    CitationsType,
    Reasoning,
    RerankModel,
    UserPrompt,
)
from nuclia.sdk.predict import AsyncNucliaPredict, NucliaPredict


def test_predict(testing_config):
    np = NucliaPredict()
    embed = np.sentence(
        text="This is my text", model="multilingual-2024-05-06", show_consumption=True
    )
    assert embed.time > 0
    assert len(embed.data) == 1024
    assert embed.consumption is not None


def test_predict_query(testing_config):
    np = NucliaPredict()
    query = np.query(
        text="Ramon, this is my text",
        semantic_model="multilingual-2024-05-06",
        token_model="multilingual",
        generative_model="chatgpt-azure-4o-mini",
        show_consumption=True,
    )
    assert query.language == "en"
    assert query.visual_llm is True
    assert query.entities and query.entities.tokens[0].text == "Ramon"
    assert query.sentence and len(query.sentence.data) == 1024
    assert query.entities.consumption is not None
    assert query.sentence.consumption is not None


def test_rag(testing_config):
    np = NucliaPredict()
    generated = np.rag(
        question="Which is the CEO of Nuclia?",
        context=[
            "Nuclia CTO is Ramon Navarro",
            "Eudald Camprubí is CEO at the same company as Ramon Navarro",
        ],
        model="chatgpt-azure-4o-mini",
        show_consumption=True,
    )
    assert "Eudald" in generated.answer
    assert generated.consumption is not None


def test_generative(testing_config):
    np = NucliaPredict()
    generated = np.generate(text="How much is 2 + 2?", model="chatgpt-azure-4o-mini")
    assert "4" in generated.answer
    assert generated.consumption is None


@pytest.mark.asyncio
async def test_generative_with_consumption(testing_config):
    np = NucliaPredict()
    generated = np.generate(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini", show_consumption=True
    )
    assert "4" in generated.answer
    assert generated.consumption is not None

    anp = AsyncNucliaPredict()
    async_generated = await anp.generate(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini", show_consumption=True
    )
    assert "4" in async_generated.answer
    assert async_generated.consumption is not None


@pytest.mark.asyncio
async def test_async_generative(testing_config):
    np = AsyncNucliaPredict()
    generated = await np.generate(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini"
    )
    assert "4" in generated.answer


def test_stream_generative(testing_config):
    np = NucliaPredict()
    found = False
    for stream in np.generate_stream(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini"
    ):
        if isinstance(stream.chunk, TextGenerativeResponse) and stream.chunk.text:
            if "4" in stream.chunk.text:
                found = True
    assert found


@pytest.mark.asyncio
async def test_async_stream_generative(testing_config):
    np = AsyncNucliaPredict()
    consumption_found = False
    found = False
    async for stream in np.generate_stream(
        text="How much is 2 + 2?", model="chatgpt-azure-4o-mini", show_consumption=True
    ):
        if isinstance(stream.chunk, TextGenerativeResponse) and stream.chunk.text:
            if "4" in stream.chunk.text:
                found = True
        elif isinstance(stream.chunk, ConsumptionGenerative):
            consumption_found = True
    assert found
    assert consumption_found


SCHEMA = {
    "name": "ClassificationReverse",
    "description": "Correctly extracted with all the required parameters with correct types",
    "parameters": {
        "$defs": {
            "Options": {
                "enum": ["SPORTS", "POLITICAL"],
                "title": "Options",
                "type": "string",
            }
        },
        "properties": {
            "title": {"default": "label", "title": "Title", "type": "string"},
            "description": {
                "default": "Define labels to classify the subject of the document",
                "title": "Description",
                "type": "string",
            },
            "document_type": {
                "description": "Type of document, SPORT example: elections, Illa, POLITICAL example: football",
                "items": {"$ref": "#/$defs/Options"},
                "title": "Document Type",
                "type": "array",
            },
        },
        "required": ["document_type"],
        "type": "object",
    },
}

TEXT = """"Many football players have existed. Messi is by far the greatest. Messi was born in Rosario, 24th of June 1987"""


@pytest.mark.asyncio
async def test_nua_parse(testing_config):
    np = AsyncNucliaPredict()
    results = await np.generate(
        text=ChatModel(
            question="",
            retrieval=False,
            user_id="Nuclia PY CLI",
            user_prompt=UserPrompt(prompt=TEXT),
            json_schema=SCHEMA,
        )
    )
    assert "SPORTS" in results.object["document_type"]


def test_nua_remi(testing_config):
    np = NucliaPredict()
    results = np.remi(
        RemiRequest(
            user_id="Nuclia PY CLI",
            question="What is the capital of France?",
            answer="Paris is the capital of france!",
            contexts=[
                "Paris is the capital of France.",
                "Berlin is the capital of Germany.",
            ],
        )
    )
    assert results.answer_relevance.score >= 4

    assert results.context_relevance[0] >= 4
    assert results.groundedness[0] >= 4

    assert results.context_relevance[1] < 2
    assert results.groundedness[1] < 2

    assert results.consumption is None


@pytest.mark.asyncio
async def test_nua_async_remi(testing_config):
    np = AsyncNucliaPredict()
    results = await np.remi(
        RemiRequest(
            user_id="Nuclia PY CLI",
            question="What is the capital of France?",
            answer="Paris is the capital of france!",
            contexts=[
                "Paris is the capital of France.",
                "Berlin is the capital of Germany.",
            ],
        ),
        show_consumption=True,
    )
    assert results.answer_relevance.score >= 4

    assert results.context_relevance[0] >= 4
    assert results.groundedness[0] >= 4

    assert results.context_relevance[1] < 2
    assert results.groundedness[1] < 2

    assert results.consumption is not None


def test_nua_rerank(testing_config):
    np = NucliaPredict()
    results = np.rerank(
        RerankModel(
            user_id="Nuclia PY CLI",
            question="What is the capital of France?",
            context={
                "1": "Paris is the capital of France.",
                "2": "Berlin is the capital of Germany.",
            },
        )
    )
    assert results.context_scores["1"] > results.context_scores["2"]
    assert results.consumption is None


@pytest.mark.asyncio
async def test_nua_rerank_with_consumption(testing_config):
    np = NucliaPredict()
    results = np.rerank(
        RerankModel(
            user_id="Nuclia PY CLI",
            question="What is the capital of France?",
            context={
                "1": "Paris is the capital of France.",
                "2": "Berlin is the capital of Germany.",
            },
        ),
        show_consumption=True,
    )
    assert results.context_scores["1"] > results.context_scores["2"]
    assert results.consumption is not None

    anp = AsyncNucliaPredict()
    async_results = await anp.rerank(
        RerankModel(
            user_id="Nuclia PY CLI",
            question="What is the capital of France?",
            context={
                "1": "Paris is the capital of France.",
                "2": "Berlin is the capital of Germany.",
            },
        ),
        show_consumption=True,
    )
    assert async_results.context_scores["1"] > async_results.context_scores["2"]
    assert async_results.consumption is not None


@pytest.mark.asyncio
async def test_generative_with_reasoning(testing_config):
    np = NucliaPredict()
    generated = np.generate(
        ChatModel(
            question=(
                "Create the simplest possible regex pattern that from the following list it matches all aws zones"
                " but not aws-il and also matches progress zone?\n\n - aws-il-central-1-1\n\n - aws-us-east-2-1\n\n"
                " - aws-europe-central-1-1\n\n - gke-prod-1\n\n - progress-proc-us-east-2-1"
            ),
            retrieval=False,
            user_id="Nuclia PY CLI",
            generative_model="chatgpt-azure-o3-mini",
            max_tokens=4000,
            reasoning=Reasoning(display=True, effort="high", budget_tokens=1024),
            user_prompt=UserPrompt(prompt="{question}"),
        ),
    )
    assert "progress" in generated.answer, generated.answer
    # Reasoning is not very consistent since the model decides when to use it
    # assert "progress" in generated.reasoning, generated.reasoning

    anp = AsyncNucliaPredict()
    async_generated = await anp.generate(
        text=ChatModel(
            question=(
                "Create the simplest possible regex pattern that from the following list it matches all aws zones"
                " but not aws-il and also matches progress zone?\n\n - aws-il-central-1-1\n\n - aws-us-east-2-1\n\n"
                " - aws-europe-central-1-1\n\n - gke-prod-1\n\n - progress-proc-us-east-2-1"
            ),
            retrieval=False,
            user_id="Nuclia PY CLI",
            generative_model="chatgpt-azure-o3-mini",
            max_tokens=4000,
            reasoning=Reasoning(display=True, effort="high", budget_tokens=1024),
            user_prompt=UserPrompt(prompt="{question}"),
        ),
    )
    assert "progress" in async_generated.answer, async_generated.answer
    # Reasoning is not very consistent since the model decides when to use it
    # assert "progress" in async_generated.reasoning, async_generated.reasoning


@pytest.fixture(scope="function")
def chat_example_md_citations() -> ChatModel:
    return ChatModel(
        question="What is Progress Agentic RAG, and what plans they provide?",
        user_id="asd",
        citations=CitationsType.LLM_FOOTNOTES,
        query_context={
            "2c6db015b370ec47abaf43aa704a16fd/f/2c6db015b370ec47abaf43aa704a16fd": "Progress Agentic RAG provides AI-powered search and knowledge management solutions.",
            "d854cbe0da054e0da186d1a722015057/l/46abb334c7afaf128e7c1e0ce877b9bb": "**Progress Agentic RAG Plans Pricing**\n* Fly: $700/month\n* Growth: $1,750/month\n* Enterprise: Contact sales",
        },
        max_tokens=1500,
    )


def test_citation_footnote_to_context(testing_config, chat_example_md_citations):
    np = NucliaPredict()
    footnote_found = False
    for stream in np.generate_stream(
        chat_example_md_citations, model="chatgpt-azure-4o-mini"
    ):
        if isinstance(stream.chunk, FootnoteCitationsGenerativeResponse):
            # Verify that footnote_to_context mapping exists
            assert stream.chunk.footnote_to_context is not None
            assert len(stream.chunk.footnote_to_context) > 0

            # Verify that the footnote IDs map to context keys from query_context
            for footnote_id, context_key in stream.chunk.footnote_to_context.items():
                assert context_key in chat_example_md_citations.query_context
                footnote_found = True

    assert footnote_found, "FootnoteCitationsGenerativeResponse chunk should be present"


@pytest.mark.asyncio
async def test_async_citation_footnote_to_context(
    testing_config, chat_example_md_citations
):
    np = AsyncNucliaPredict()
    footnote_found = False
    async for stream in np.generate_stream(
        chat_example_md_citations, model="chatgpt-azure-4o-mini"
    ):
        if isinstance(stream.chunk, FootnoteCitationsGenerativeResponse):
            # Verify that footnote_to_context mapping exists
            assert stream.chunk.footnote_to_context is not None
            assert len(stream.chunk.footnote_to_context) > 0

            # Verify that the footnote IDs map to context keys from query_context
            for footnote_id, context_key in stream.chunk.footnote_to_context.items():
                assert context_key in chat_example_md_citations.query_context
                footnote_found = True

    assert footnote_found, "FootnoteCitationsGenerativeResponse chunk should be present"



import os
import tempfile

import pytest

from nuclia import BASE_DOMAIN
from nuclia.config import reset_config_file, set_config_file
from nuclia.sdk.auth import NucliaAuth

IS_PROD = False
if BASE_DOMAIN == "stashify.cloud":
    TESTING_ACCOUNT_SLUG = "eric-cicd"
    TESTING_KBID = "3fc11430-e1d5-45c7-86de-86d8efdd2cac"
    TESTING_KB = "https://europe-1.stashify.cloud/api/v1/kb/" + TESTING_KBID
    TESTING_AGENT_ID = "3d9ab1b7-da33-4709-bc95-03d30c989f56"
    TESTING_AGENT = "https://europe-1.stashify.cloud/api/v1/kb/" + TESTING_AGENT_ID
else:
    IS_PROD = True
    TESTING_ACCOUNT_SLUG = "nuclia"
    TESTING_KBID = "18ab102c-a7db-4a35-b894-c20422b3b9f0"
    TESTING_KB = "https://europe-1.rag.progress.cloud/api/v1/kb/" + TESTING_KBID
    TESTING_AGENT_ID = "6ce34422-ee0d-4077-9bc5-3845fe4d34f4"
    TESTING_AGENT = "https://europe-1.rag.progress.cloud/api/v1/kb/" + TESTING_AGENT_ID


@pytest.fixture(scope="module")
def testing_user():
    return os.environ.get("GA_TESTING_TOKEN")


@pytest.fixture(scope="module")
def testing_nua():
    return os.environ.get("GA_TESTING_NUA")


@pytest.fixture(scope="module")
def testing_kb():
    return os.environ.get("GA_TESTING_SERVICE_TOKEN")


@pytest.fixture(scope="module")
def testing_agent():
    return os.environ.get("GA_TESTING_AGENT_TOKEN")


@pytest.fixture(scope="module")
def testing_config(testing_kb, testing_nua, testing_user, testing_agent):
    os.environ["TESTING"] = "True"
    with tempfile.NamedTemporaryFile() as temp_file:
        temp_file.write(b"{}")
        temp_file.flush()
        set_config_file(temp_file.name)
        nuclia_auth = NucliaAuth()
        nuclia_auth.set_user_token(testing_user)
        nuclia_auth.kb(TESTING_KB, testing_kb)
        nuclia_auth.agent(TESTING_AGENT, testing_agent)
        client_id = nuclia_auth.nua(testing_nua)
        assert client_id
        nuclia_auth._config.set_default_kb(TESTING_KBID)
        nuclia_auth._config.set_default_nua(client_id)
        nuclia_auth._config.set_default_agent(TESTING_AGENT_ID)
        yield
        reset_config_file()


@pytest.fixture(autouse=True)
def cleanup_auth():
    # sdk stores the client on DATA, so when running async tests, that gets tied to a function scope loop
    # and breaks all consequent tests...
    from nuclia import data

    data.DATA.async_auth = None



from typing import Type, Union

import pytest
from nuclia_models.agent.interaction import AnswerOperation, AragAnswer

from nuclia.sdk.agent import AsyncNucliaAgent, NucliaAgent
from nuclia.tests.utils import maybe_async_iterate


@pytest.mark.parametrize(
    "agent_klass",
    [NucliaAgent, AsyncNucliaAgent],
)
async def test_interact(
    testing_config,
    agent_klass: Type[Union[NucliaAgent, AsyncNucliaAgent]],
):
    n_agent = agent_klass()
    responses: list[AragAnswer] = []
    async for message in maybe_async_iterate(
        n_agent.interact(
            session_uuid="ephemeral",
            question="What is Eric known for?",
            headers={"X-Custom-Header": "value"},
        )
    ):
        responses.append(message)
    assert len(responses) > 0
    assert responses[0].operation == AnswerOperation.START

    assert responses[1].operation == AnswerOperation.ANSWER
    assert responses[1].step and responses[1].step.module == "rephrase"

    assert responses[2].operation == AnswerOperation.ANSWER
    assert responses[2].step and responses[2].step.module == "basic_ask"

    assert responses[-3].operation == AnswerOperation.ANSWER
    assert responses[-3].step and responses[-3].step.module == "remi"

    assert responses[-2].operation == AnswerOperation.ANSWER
    assert responses[-2].answer and "humor" in responses[-2].answer.lower()

    assert responses[-1].operation == AnswerOperation.DONE



from contextlib import asynccontextmanager, contextmanager
from typing import (
    AsyncGenerator,
    AsyncIterator,
    Generator,
    Optional,
    Type,
    TypeVar,
    Union,
    overload,
)

from httpx import AsyncClient, Client
from nuclia_models.agent.api import SessionData
from nuclia_models.agent.interaction import (
    AnswerOperation,
    AragAnswer,
    InteractionOperation,
    InteractionRequest,
    UserToAgentInteraction,
)
from nuclia_models.common.formats import TextFormat
from nucliadb_models.resource import Resource, ResourceList
from nucliadb_models.writer import ResourceCreated
from pydantic import BaseModel
from websockets.asyncio.client import ClientConnection as AsyncClientConnection
from websockets.asyncio.client import connect as async_connect
from websockets.sync.client import connect

from nuclia.exceptions import RaoAPIException
from nuclia.lib.utils import (
    USER_AGENT,
    build_httpx_async_client,
    build_httpx_client,
)

ConvertType = TypeVar("ConvertType", bound=BaseModel)


class BaseAgentClient:
    """Base class containing common logic for Agent clients."""

    region: str
    agent_id: str
    account_id: str
    headers: dict[str, str]
    user_token: Optional[str]
    url: str
    ws_url: str

    def __init__(
        self,
        region: str,
        agent_id: str,
        account_id: str,
        api_key: Optional[str] = None,
        user_token: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
    ):
        # Import here otherwise we might use outdated values
        from nuclia import REGIONAL

        self.region = region
        self.agent_id = agent_id
        self.account_id = account_id
        self.user_token = user_token

        # Build URL from region
        if "http" in region:
            self.url = region.strip("/")
        else:
            self.url = REGIONAL.format(region=region).strip("/")

        # Build WebSocket URL
        self.ws_url = self.url.replace("https://", "wss://")

        # Build headers
        self.headers = {"User-Agent": USER_AGENT}
        if api_key is not None:
            self.headers["X-NUCLIA-SERVICEACCOUNT"] = f"Bearer {api_key}"
        elif user_token is not None:
            self.headers["Authorization"] = f"Bearer {user_token}"
        if headers is not None:
            self.headers = self.headers | headers


class AgentClient(BaseAgentClient):
    """Synchronous Agent client."""

    http_client: Client

    def __init__(
        self,
        region: str,
        agent_id: str,
        account_id: str,
        api_key: Optional[str] = None,
        user_token: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
    ):
        super().__init__(region, agent_id, account_id, api_key, user_token, headers)
        self.http_client = build_httpx_client(headers=self.headers, base_url=self.url)

    @overload
    def _request(
        self,
        method: str,
        url: str,
        output: None = None,
        payload: Union[dict, BaseModel, None] = None,
        params: Optional[dict] = None,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 60,
    ) -> dict: ...

    @overload
    def _request(
        self,
        method: str,
        url: str,
        output: Type[ConvertType],
        payload: Union[dict, BaseModel, None] = None,
        params: Optional[dict] = None,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 60,
    ) -> ConvertType: ...

    def _request(
        self,
        method: str,
        url: str,
        output: Optional[Type[ConvertType]] = None,
        payload: Union[dict, BaseModel, None] = None,
        params: Optional[dict] = None,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 60,
    ) -> Union[ConvertType, dict]:
        if isinstance(payload, BaseModel):
            payload = payload.model_dump(mode="json")
        resp = self.http_client.request(
            method,
            url,
            json=payload,
            timeout=timeout,
            headers=extra_headers,
            params=params,
        )
        if resp.status_code != 200 and resp.status_code != 201:
            raise RaoAPIException(code=resp.status_code, detail=resp.content.decode())
        if output:
            try:
                data = output.model_validate(resp.json())
            except Exception:
                data = output.model_validate(resp.content)
        else:
            data = resp.json()
        return data

    def new_session(self, name: str) -> str:
        return self.create_session(
            SessionData(
                slug=name.replace(" ", "-").lower(),
                name=name,
                summary="",
                data="",
                format=TextFormat.PLAIN,
            )
        ).uuid

    def create_session(self, session_data: SessionData) -> ResourceCreated:
        return self._request(
            "POST",
            f"/api/v1/agent/{self.agent_id}/sessions",
            output=ResourceCreated,
            payload=session_data,
        )

    def delete_session(self, session_uuid: str) -> None:
        self._request(
            "DELETE", f"/api/v1/kb/{self.agent_id}/agent/session/{session_uuid}"
        )

    def get_sessions(self, page: int = 0, page_size=20) -> ResourceList:
        return self._request(
            "GET",
            f"/api/v1/kb/{self.agent_id}/agent/sessions",
            output=ResourceList,
            params={"page": page, "page_size": page_size},
        )

    def get_session(self, session_uuid: str) -> Resource:
        return self._request(
            "GET",
            f"/api/v1/kb/{self.agent_id}/agent/session/{session_uuid}",
            output=Resource,
        )

    def get_ephemeral_token(self, session_uuid: str) -> str:
        resp = self._request(
            "POST",
            f"/api/v1/account/{self.account_id}/kb/{self.agent_id}/ephemeral_tokens",
            payload={"agent_session": session_uuid},
        )
        return resp["token"]

    @contextmanager
    def websocket(self, session_uuid: str):
        eph_token = self.get_ephemeral_token(session_uuid)
        with connect(
            f"{self.ws_url}/api/v1/agent/{self.agent_id}/session/{session_uuid}/ws?eph-token={eph_token}"
        ) as websocket:
            yield websocket

    def interact(
        self, session_uuid: str, question: str, headers: Optional[dict[str, str]] = None
    ) -> Generator[AragAnswer, Optional[str], None]:
        message = InteractionRequest(
            question=question,
            headers=headers or {},
            operation=InteractionOperation.QUESTION,
        ).model_dump_json()
        with self.websocket(session_uuid) as websocket:
            websocket.send(message)
            user_response = None
            for data in websocket:
                response = AragAnswer.model_validate_json(data)

                # If we need a user response, yield the response and wait for input
                if response.operation == AnswerOperation.AGENT_REQUEST:
                    user_response = yield response
                    if user_response and response.agent_request:
                        # Send user response back to agent
                        user_msg = UserToAgentInteraction(
                            request_id=response.agent_request,
                            response=user_response,
                        ).model_dump_json()
                        websocket.send(user_msg)
                else:
                    yield response


class AsyncAgentClient(BaseAgentClient):
    """Asynchronous Agent client."""

    http_client: AsyncClient

    def __init__(
        self,
        region: str,
        agent_id: str,
        account_id: str,
        api_key: Optional[str] = None,
        user_token: Optional[str] = None,
        headers: Optional[dict[str, str]] = None,
    ):
        super().__init__(region, agent_id, account_id, api_key, user_token, headers)
        self.http_client = build_httpx_async_client(
            headers=self.headers, base_url=self.url
        )

    @overload
    async def _request(
        self,
        method: str,
        url: str,
        output: None = None,
        payload: Union[dict, BaseModel, None] = None,
        params: Optional[dict] = None,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 60,
    ) -> dict: ...

    @overload
    async def _request(
        self,
        method: str,
        url: str,
        output: Type[ConvertType],
        payload: Union[dict, BaseModel, None] = None,
        params: Optional[dict] = None,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 60,
    ) -> ConvertType: ...

    async def _request(
        self,
        method: str,
        url: str,
        output: Optional[Type[ConvertType]] = None,
        payload: Union[dict, BaseModel, None] = None,
        params: Optional[dict] = None,
        extra_headers: Optional[dict[str, str]] = None,
        timeout: int = 60,
    ) -> Union[ConvertType, dict]:
        if isinstance(payload, BaseModel):
            payload = payload.model_dump(mode="json")
        resp = await self.http_client.request(
            method,
            url,
            json=payload,
            timeout=timeout,
            headers=extra_headers,
            params=params,
        )
        if resp.status_code != 200 and resp.status_code != 201:
            raise RaoAPIException(code=resp.status_code, detail=resp.content.decode())
        if output:
            try:
                data = output.model_validate(resp.json())
            except Exception:
                data = output.model_validate(resp.content)
        else:
            data = resp.json()
        return data

    async def new_session(self, name: str) -> str:
        return (
            await self.create_session(
                SessionData(
                    slug=name.replace(" ", "-").lower(),
                    name=name,
                    summary="",
                    data="",
                    format=TextFormat.PLAIN,
                )
            )
        ).uuid

    async def create_session(self, session_data: SessionData) -> ResourceCreated:
        return await self._request(
            "POST",
            f"/api/v1/agent/{self.agent_id}/sessions",
            output=ResourceCreated,
            payload=session_data,
        )

    async def delete_session(self, session_uuid: str) -> None:
        await self._request(
            "DELETE", f"/api/v1/kb/{self.agent_id}/agent/session/{session_uuid}"
        )

    async def get_sessions(self, page: int = 0, page_size=20) -> ResourceList:
        return await self._request(
            "GET",
            f"/api/v1/kb/{self.agent_id}/agent/sessions",
            output=ResourceList,
            params={"page": page, "page_size": page_size},
        )

    async def get_session(self, session_uuid: str) -> Resource:
        return await self._request(
            "GET",
            f"/api/v1/kb/{self.agent_id}/agent/session/{session_uuid}",
            output=Resource,
        )

    async def get_ephemeral_token(self, session_uuid: str) -> str:
        resp = await self._request(
            "POST",
            f"/api/v1/account/{self.account_id}/kb/{self.agent_id}/ephemeral_tokens",
            payload={"agent_session": session_uuid},
        )
        return resp["token"]

    @asynccontextmanager
    async def websocket(
        self, session_uuid: str
    ) -> AsyncGenerator[AsyncClientConnection, None]:
        eph_token = await self.get_ephemeral_token(session_uuid)
        async with async_connect(
            f"{self.ws_url}/api/v1/agent/{self.agent_id}/session/{session_uuid}/ws?eph-token={eph_token}",
        ) as websocket:
            yield websocket

    async def interact(
        self, session_uuid: str, question: str, headers: Optional[dict[str, str]] = None
    ) -> AsyncIterator[AragAnswer]:
        message = InteractionRequest(
            question=question,
            headers=headers or {},
            operation=InteractionOperation.QUESTION,
        ).model_dump_json()
        async with self.websocket(session_uuid) as websocket:
            await websocket.send(message)
            async for data in websocket:
                response = AragAnswer.model_validate_json(data)
                yield response



import os
from typing import Type, Union

import pytest
from nucliadb_sdk.v2.exceptions import NotFoundError

from nuclia.sdk.kb import AsyncNucliaUpload, NucliaUpload
from nuclia.sdk.resource import NucliaResource
from nuclia.tests.utils import maybe_await


@pytest.mark.parametrize(
    "nupload_klass",
    [NucliaUpload, AsyncNucliaUpload],
)
async def test_upload_without_payload(
    testing_config,
    nupload_klass: Union[Type[NucliaUpload], Type[AsyncNucliaUpload]],
):
    nupload = nupload_klass()
    nresource = NucliaResource()
    slug = "upload_without_payload"

    try:
        nresource.delete(slug=slug)
    except NotFoundError:
        pass

    rid = await maybe_await(
        nupload.file(
            path=f"{os.path.dirname(__file__)}/test_upload.py",
            slug=slug,
        )
    )
    res = nresource.get(rid=rid)
    assert res.title == slug


@pytest.mark.parametrize(
    "nupload_klass",
    [NucliaUpload, AsyncNucliaUpload],
)
async def test_upload_with_labels(
    testing_config,
    nupload_klass: Union[Type[NucliaUpload], Type[AsyncNucliaUpload]],
):
    nupload = nupload_klass()
    nresource = NucliaResource()
    slug = "upload_with_labels"

    try:
        nresource.delete(slug=slug)
    except NotFoundError:
        pass

    rid = await maybe_await(
        nupload.file(
            path=f"{os.path.dirname(__file__)}/test_upload.py",
            slug=slug,
            usermetadata={"classifications": [{"labelset": "a", "label": "b"}]},
        )
    )
    res = nresource.get(rid=rid)
    assert res.title == slug
    labels = res.usermetadata.classifications
    assert labels
    assert labels[0].labelset == "a"
    assert labels[0].label == "b"



import asyncio
import os
import tempfile
import time
from datetime import datetime
from typing import Any, List, Optional, Union

from deprecated import deprecated
from nucliadb_models import Notification
from nucliadb_models.labels import KnowledgeBoxLabels, Label, LabelSet, LabelSetKind
from nucliadb_models.resource import ResourceList
from nucliadb_models.search import SummarizeRequest, SummaryKind
from nucliadb_sdk import exceptions

from nuclia.data import get_async_auth, get_async_client, get_auth, get_client
from nuclia.decorators import kb
from nuclia.lib.kb import AsyncNucliaDBClient, NucliaDBClient
from nuclia.lib.models import GraphRelation, get_relation
from nuclia.lib.nua_responses import SummarizedModel
from nuclia.sdk.auth import AsyncNucliaAuth, NucliaAuth
from nuclia.sdk.export_import import (
    AsyncNucliaExports,
    AsyncNucliaImports,
    NucliaExports,
    NucliaImports,
)
from nuclia.sdk.extract_strategy import (
    AsyncNucliaExtractStrategy,
    NucliaExtractStrategy,
)
from nuclia.sdk.logger import logger
from nuclia.sdk.logs import AsyncNucliaLogs, NucliaLogs
from nuclia.sdk.remi import AsyncNucliaRemi, NucliaRemi
from nuclia.sdk.resource import AsyncNucliaResource, NucliaResource
from nuclia.sdk.search import AsyncNucliaSearch, NucliaSearch
from nuclia.sdk.split_strategy import (
    AsyncNucliaSplitStrategy,
    NucliaSplitStrategy,
)
from nuclia.sdk.task import AsyncNucliaTask, NucliaTask
from nuclia.sdk.upload import AsyncNucliaUpload, NucliaUpload


class NucliaKB:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    def __init__(self):
        self.upload = NucliaUpload()
        self.search = NucliaSearch()
        self.resource = NucliaResource()
        self.exports = NucliaExports()
        self.imports = NucliaImports()
        self.logs = NucliaLogs()
        self.task = NucliaTask()
        self.remi = NucliaRemi()
        self.extract_strategies = NucliaExtractStrategy()
        self.split_strategies = NucliaSplitStrategy()

    @kb
    def list(
        self, page: Optional[int] = None, size: Optional[int] = None, **kwargs
    ) -> ResourceList:
        ndb: NucliaDBClient = kwargs["ndb"]
        query_params = {}
        if page:
            query_params["page"] = f"{page}"
        if size:
            query_params["size"] = f"{size}"
        data: ResourceList = ndb.ndb.list_resources(
            kbid=ndb.kbid, query_params=query_params
        )
        return data

    @kb
    def get_labelset(
        self,
        *,
        labelset: str,
        **kwargs,
    ) -> LabelSet:
        ndb: NucliaDBClient = kwargs["ndb"]
        return ndb.ndb.get_labelset(kbid=ndb.kbid, labelset=labelset)

    @kb
    @deprecated(version="5.0.0", reason="You should use set_labelset")
    def add_labelset(
        self,
        *,
        labelset: str,
        kind: LabelSetKind = LabelSetKind.RESOURCES,
        multiple: bool = True,
        title: Optional[str] = None,
        color: Optional[str] = None,
        labels: Optional[List[str]] = None,
        **kwargs,
    ):
        self.set_labelset(
            labelset=labelset,
            kind=kind,
            multiple=multiple,
            title=title,
            color=color,
            labels=labels,
            **kwargs,
        )

    @kb
    def set_labelset(
        self,
        *,
        labelset: str,
        kind: LabelSetKind = LabelSetKind.RESOURCES,
        multiple: bool = True,
        title: Optional[str] = None,
        color: Optional[str] = None,
        labels: Optional[List[str]] = None,
        **kwargs,
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        if labels is None:
            labels_list = []
        else:
            labels_list = [Label(title=label) for label in labels]

        labelset_obj = LabelSet(
            title=title if title is not None else labelset,
            color=color if color is not None else "blue",
            labels=labels_list,
            kind=[kind],
            multiple=multiple,
        )

        ndb.ndb.set_labelset(kbid=ndb.kbid, labelset=labelset, content=labelset_obj)

    @kb
    def list_labelsets(self, **kwargs) -> KnowledgeBoxLabels:
        ndb: NucliaDBClient = kwargs["ndb"]
        data: KnowledgeBoxLabels = ndb.ndb.get_labelsets(kbid=ndb.kbid)
        return data

    @kb
    def del_labelset(self, *, labelset: str, **kwargs):
        ndb: NucliaDBClient = kwargs["ndb"]
        ndb.ndb.delete_labelset(kbid=ndb.kbid, labelset=labelset)

    @kb
    def add_label(
        self,
        *,
        labelset: str,
        label: str,
        labelset_kind: Optional[LabelSetKind] = None,
        **kwargs,
    ):
        self.add_labels(
            labelset=labelset,
            labels=[label],
            labelset_kind=labelset_kind,
            **kwargs,
        )

    @kb
    def add_labels(
        self,
        *,
        labelset: str,
        labels: List[str],
        labelset_kind: Optional[LabelSetKind] = None,
        **kwargs,
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        existing = False
        try:
            labelset_obj: LabelSet = ndb.ndb.get_labelset(
                kbid=ndb.kbid, labelset=labelset
            )
            existing = True
        except exceptions.NotFoundError:
            pass
        if not existing:
            if labelset_kind is None:
                raise ValueError("Labelset kind must be defined for a new labelset")
            else:
                labelset_obj = LabelSet(
                    title=labelset,
                    color="blue",
                    labels=[],
                    kind=[labelset_kind],
                    multiple=True,
                )
        existing_labels = [x.title for x in labelset_obj.labels]
        for label in labels:
            if label in existing_labels:
                continue
            label_obj = Label(title=label)
            labelset_obj.labels.append(label_obj)
        ndb.ndb.set_labelset(
            kbid=ndb.kbid,
            labelset=labelset,
            content=labelset_obj,
        )

    @kb
    def del_label(self, *, labelset: str, label: str, **kwargs):
        ndb: NucliaDBClient = kwargs["ndb"]
        labelset_obj: LabelSet = ndb.ndb.get_labelset(kbid=ndb.kbid, labelset=labelset)
        label_to_delete = next(x for x in labelset_obj.labels if x.title == label)
        labelset_obj.labels.remove(label_to_delete)
        ndb.ndb.set_labelset(kbid=ndb.kbid, labelset=labelset, content=labelset_obj)

    @kb
    def get_graph(
        self,
        uid: Optional[str] = None,
        slug: Optional[str] = None,
        **kwargs,
    ):
        res = self.resource.get(
            ndb=kwargs["ndb"], rid=uid, slug=slug, show=["basic", "relations"]
        )
        return res.usermetadata.relations

    @kb
    def add_graph(
        self,
        graph: List[Union[GraphRelation, dict]],
        slug: Optional[str] = None,
        **kwargs,
    ):
        kw = {
            "ndb": kwargs["ndb"],
            "usermetadata": {
                "relations": [
                    get_relation(relation).to_relation() for relation in graph
                ]
            },
        }
        if slug:
            kw["slug"] = slug
        self.resource.create(**kw)

    @kb
    def update_graph(
        self,
        graph: List[Union[GraphRelation, dict]],
        uid: Optional[str] = None,
        slug: Optional[str] = None,
        override: Optional[bool] = False,
        **kwargs,
    ):
        new_relations = [get_relation(relation).to_relation() for relation in graph]
        if not override:
            relations = self.get_graph(uid=uid, slug=slug, **kwargs)
            relations.extend(new_relations)
        else:
            relations = new_relations
        self.resource.update(
            ndb=kwargs["ndb"], rid=uid, slug=slug, usermetadata={"relations": relations}
        )

    @kb
    def delete_graph(
        self,
        uid: Optional[str] = None,
        slug: Optional[str] = None,
        **kwargs,
    ):
        self.resource.delete(ndb=kwargs["ndb"], rid=uid, slug=slug)

    @kb
    def update_configuration(
        self,
        *,
        semantic_model: Optional[str] = None,
        generative_model: Optional[str] = None,
        ner_model: Optional[str] = None,
        anonymization_model: Optional[str] = None,
        visual_labeling: Optional[str] = None,
        **kwargs,
    ):
        ndb: NucliaDBClient = kwargs.pop("ndb")

        content = {}
        if semantic_model is not None:
            content["semantic_model"] = semantic_model
        if generative_model is not None:
            content["generative_model"] = generative_model
        if ner_model is not None:
            content["ner_model"] = ner_model
        if anonymization_model is not None:
            content["anonymization_model"] = anonymization_model
        if visual_labeling is not None:
            content["visual_labeling"] = visual_labeling

        ndb.ndb.update_configuration(
            kbid=ndb.kbid,
            content=content,
        )

    @kb
    def get_configuration(
        self,
        **kwargs,
    ) -> Any:
        ndb: NucliaDBClient = kwargs["ndb"]
        return ndb.ndb.get_configuration(
            kbid=ndb.kbid,
        )

    @kb
    def summarize(
        self, *, resources: List[str], show_consumption: bool = False, **kwargs
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        return ndb.ndb.summarize(
            kbid=ndb.kbid,
            resources=resources,
            headers={"X-Show-Consumption": str(show_consumption).lower()},
        )

    @kb
    def notifications(self, **kwargs):
        ndb: NucliaDBClient = kwargs["ndb"]
        response = ndb.notifications()
        for notification in response.iter_lines():
            yield Notification.model_validate_json(notification)

    @kb
    def copy(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        destination: str,
        override: Optional[bool] = False,
        **kwargs,
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        if rid:
            res = ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid,
                rid=rid,
                query_params={
                    "show": ["basic", "origin", "extra", "values", "security"]
                },
            )
        elif slug:
            res = ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid,
                slug=slug,
                query_params={
                    "show": ["basic", "origin", "extra", "values", "security"]
                },
            )
        else:
            raise ValueError("Either rid or slug must be provided")
        data = {
            "kbid": destination,
            "slug": res.slug,
            "title": res.title,
            "summary": res.summary,
            "icon": res.icon,
            "origin": res.origin,
            "extra": res.extra,
            "usermetadata": res.usermetadata,
            "fieldmetadata": res.fieldmetadata,
            "security": res.security,
        }
        files_to_upload = []
        if res.data.conversations:
            data["conversations"] = dict(
                zip(
                    res.data.conversations.keys(),
                    [dict(v.value) for v in res.data.conversations.values()],
                )
            )
        if res.data.links:
            data["links"] = dict(
                zip(
                    res.data.links.keys(),
                    [dict(v.value) for v in res.data.links.values()],
                )
            )
        if res.data.texts:
            data["texts"] = dict(
                zip(
                    res.data.texts.keys(),
                    [dict(v.value) for v in res.data.texts.values()],
                )
            )
        if res.data.files:
            remote_files = {}
            for file_id, file in res.data.files.items():
                if file.value.external:
                    remote_files[file_id] = file.value
                else:
                    files_to_upload.append({"id": file_id, "data": file.value})
        destination_kb = get_client(destination)
        if override:
            try:
                self.resource.delete(
                    ndb=destination_kb,
                    slug=res.slug,
                )
            except exceptions.NotFoundError:
                pass
        failed = False
        try:
            uuid = self.resource.create(ndb=destination_kb, **data)
        except exceptions.RateLimitError as e:
            failed = True
            delay = 5
            if e.try_after:
                delay = round(e.try_after - datetime.utcnow().timestamp())
            logger.warning(
                f"Backpressure error while copying resource, retrying in {delay} seconds"
            )
            time.sleep(delay)
            self.copy(
                rid=rid, slug=slug, destination=destination, override=override, **kwargs
            )

        if not failed:
            with tempfile.TemporaryDirectory() as tmpdirname:
                for file_data in files_to_upload:
                    file_path = os.path.join(
                        tmpdirname, file_data["data"].file.filename
                    )
                    self.resource.download_file(
                        rid=res.id, file_id=file_data["id"], output=file_path, **kwargs
                    )
                    self.upload.file(path=file_path, ndb=destination_kb, rid=uuid)

    @kb
    def copy_all(
        self,
        *,
        destination: str,
        page=0,
        override: Optional[bool] = False,
        filters: Optional[List[str]] = None,
        **kwargs,
    ):
        ndb: NucliaDBClient = kwargs["ndb"]
        if filters is None:
            batch = self.list(ndb=ndb, page=page)
            resources = batch.resources
            is_last = batch.pagination.last
        else:
            batch = self.search.find(
                ndb=ndb, filters=filters, page=page, fields=["a/title"]
            )
            resources = batch.resources.values()
            is_last = batch.next_page
        for res in resources:
            try:
                logger.info(f"Copying resource {res.id}")
                self.copy(
                    rid=res.id, destination=destination, override=override, **kwargs
                )
            except exceptions.ConflictError:
                logger.info(f"Resource {res.id} already exists in destination KB")
        if not is_last:
            self.copy_all(
                destination=destination, page=page + 1, override=override, **kwargs
            )


class AsyncNucliaKB:
    @property
    def _auth(self) -> AsyncNucliaAuth:
        auth = get_async_auth()
        return auth

    def __init__(self):
        self.upload = AsyncNucliaUpload()
        self.search = AsyncNucliaSearch()
        self.resource = AsyncNucliaResource()
        self.exports = AsyncNucliaExports()
        self.imports = AsyncNucliaImports()
        self.logs = AsyncNucliaLogs()
        self.task = AsyncNucliaTask()
        self.remi = AsyncNucliaRemi()
        self.extract_strategies = AsyncNucliaExtractStrategy()
        self.split_strategies = AsyncNucliaSplitStrategy()

    @kb
    async def list(self, **kwargs) -> ResourceList:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        del kwargs["ndb"]
        data: ResourceList = await ndb.ndb.list_resources(
            kbid=ndb.kbid, query_params=kwargs
        )
        return data

    @kb
    async def get_labelset(
        self,
        *,
        labelset: str,
        **kwargs,
    ) -> LabelSet:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        return await ndb.ndb.get_labelset(kbid=ndb.kbid, labelset=labelset)

    @kb
    @deprecated(version="5.0.0", reason="You should use set_labelset")
    async def add_labelset(
        self,
        *,
        labelset: str,
        kind: LabelSetKind = LabelSetKind.RESOURCES,
        multiple: bool = True,
        title: Optional[str] = None,
        color: Optional[str] = None,
        labels: Optional[List[str]] = None,
        **kwargs,
    ):
        await self.set_labelset(
            labelset=labelset,
            kind=kind,
            multiple=multiple,
            title=title,
            color=color,
            labels=labels,
            **kwargs,
        )

    @kb
    async def set_labelset(
        self,
        *,
        labelset: str,
        kind: LabelSetKind = LabelSetKind.RESOURCES,
        multiple: bool = True,
        title: Optional[str] = None,
        color: Optional[str] = None,
        labels: Optional[List[str]] = None,
        **kwargs,
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if labels is None:
            labels_list = []
        else:
            labels_list = [Label(title=label) for label in labels]

        labelset_obj = LabelSet(
            title=title if title is not None else labelset,
            color=color if color is not None else "blue",
            labels=labels_list,
            kind=[kind],
            multiple=multiple,
        )

        await ndb.ndb.set_labelset(
            kbid=ndb.kbid, labelset=labelset, content=labelset_obj
        )

    @kb
    async def list_labelsets(self, **kwargs) -> KnowledgeBoxLabels:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        data: KnowledgeBoxLabels = await ndb.ndb.get_labelsets(kbid=ndb.kbid)
        return data

    @kb
    async def del_labelset(self, *, labelset: str, **kwargs):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        await ndb.ndb.delete_labelset(kbid=ndb.kbid, labelset=labelset)

    @kb
    async def add_label(
        self,
        *,
        labelset: str,
        label: str,
        labelset_kind: Optional[LabelSetKind] = None,
        **kwargs,
    ):
        await self.add_labels(
            labelset=labelset,
            labels=[label],
            labelset_kind=labelset_kind,
            **kwargs,
        )

    @kb
    async def add_labels(
        self,
        *,
        labelset: str,
        labels: List[str],
        labelset_kind: Optional[LabelSetKind] = None,
        **kwargs,
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        existing = False
        try:
            labelset_obj: LabelSet = await ndb.ndb.get_labelset(
                kbid=ndb.kbid, labelset=labelset
            )
            existing = True
        except exceptions.NotFoundError:
            pass
        if not existing:
            if labelset_kind is None:
                raise ValueError("Labelset kind must be defined for a new labelset")
            else:
                labelset_obj = LabelSet(
                    title=labelset,
                    color="blue",
                    labels=[],
                    kind=[labelset_kind],
                    multiple=True,
                )
        existing_labels = [x.title for x in labelset_obj.labels]
        for label in labels:
            if label in existing_labels:
                continue
            label_obj = Label(title=label)
            labelset_obj.labels.append(label_obj)
        await ndb.ndb.set_labelset(
            kbid=ndb.kbid,
            labelset=labelset,
            content=labelset_obj,
        )

    @kb
    async def del_label(self, *, labelset: str, label: str, **kwargs):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        labelset_obj: LabelSet = await ndb.ndb.get_labelset(
            kbid=ndb.kbid, labelset=labelset
        )
        label_to_delete = next(x for x in labelset_obj.labels if x.title == label)
        labelset_obj.labels.remove(label_to_delete)
        await ndb.ndb.set_labelset(
            kbid=ndb.kbid, labelset=labelset, content=labelset_obj
        )

    @kb
    async def get_graph(
        self,
        uid: Optional[str] = None,
        slug: Optional[str] = None,
        **kwargs,
    ):
        res = await self.resource.get(
            ndb=kwargs["ndb"], rid=uid, slug=slug, show=["basic", "relations"]
        )
        return res.usermetadata.relations

    @kb
    async def add_graph(
        self,
        graph: List[Union[GraphRelation, dict]],
        slug: Optional[str] = None,
        **kwargs,
    ):
        kw = {
            "ndb": kwargs["ndb"],
            "usermetadata": {
                "relations": [
                    get_relation(relation).to_relation() for relation in graph
                ]
            },
        }
        if slug:
            kw["slug"] = slug
        await self.resource.create(**kw)

    @kb
    async def update_graph(
        self,
        graph: List[Union[GraphRelation, dict]],
        uid: Optional[str] = None,
        slug: Optional[str] = None,
        override: Optional[bool] = False,
        **kwargs,
    ):
        new_relations = [get_relation(relation).to_relation() for relation in graph]
        if not override:
            relations = await self.get_graph(uid=uid, slug=slug, **kwargs)
            relations.extend(new_relations)
        else:
            relations = new_relations
        await self.resource.update(
            ndb=kwargs["ndb"], rid=uid, slug=slug, usermetadata={"relations": relations}
        )

    @kb
    async def delete_graph(
        self,
        uid: Optional[str] = None,
        slug: Optional[str] = None,
        **kwargs,
    ):
        await self.resource.delete(ndb=kwargs["ndb"], rid=uid, slug=slug)

    @kb
    async def update_configuration(
        self,
        *,
        semantic_model: Optional[str] = None,
        generative_model: Optional[str] = None,
        ner_model: Optional[str] = None,
        anonymization_model: Optional[str] = None,
        visual_labeling: Optional[str] = None,
        **kwargs,
    ):
        ndb: AsyncNucliaDBClient = kwargs.pop("ndb")

        content = {}
        if semantic_model is not None:
            content["semantic_model"] = semantic_model
        if generative_model is not None:
            content["generative_model"] = generative_model
        if ner_model is not None:
            content["ner_model"] = ner_model
        if anonymization_model is not None:
            content["anonymization_model"] = anonymization_model
        if visual_labeling is not None:
            content["visual_labeling"] = visual_labeling

        await ndb.ndb.update_configuration(
            kbid=ndb.kbid,
            content=content,
        )

    @kb
    async def get_configuration(
        self,
        **kwargs,
    ) -> Any:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        return await ndb.ndb.get_configuration(
            kbid=ndb.kbid,
        )

    @kb
    async def summarize(
        self,
        *,
        resources: List[str],
        generative_model: Optional[str] = None,
        summary_kind: Optional[str] = None,
        show_consumption: bool = False,
        timeout: int = 1000,
        **kwargs,
    ) -> SummarizedModel:
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        resp = await ndb.summarize(
            SummarizeRequest(
                resources=resources,
                generative_model=generative_model,
                summary_kind=SummaryKind(summary_kind),
            ),
            extra_headers={"X-Show-Consumption": str(show_consumption).lower()},
            timeout=timeout,
        )
        return SummarizedModel.model_validate(resp.json())

    @kb
    async def notifications(self, **kwargs):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        response = await ndb.notifications()
        async for notification in response.aiter_lines():
            yield Notification.model_validate_json(notification)

    @kb
    async def copy(
        self,
        *,
        rid: Optional[str] = None,
        slug: Optional[str] = None,
        destination: str,
        **kwargs,
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if rid:
            res = await ndb.ndb.get_resource_by_id(
                kbid=ndb.kbid,
                rid=rid,
                query_params={
                    "show": ["basic", "origin", "extra", "values", "security"]
                },
            )
        elif slug:
            res = await ndb.ndb.get_resource_by_slug(
                kbid=ndb.kbid,
                slug=slug,
                query_params={
                    "show": ["basic", "origin", "extra", "values", "security"]
                },
            )
        else:
            raise ValueError("Either rid or slug must be provided")
        data = {
            "kbid": destination,
            "slug": res.slug,
            "title": res.title,
            "summary": res.summary,
            "icon": res.icon,
            "origin": res.origin,
            "extra": res.extra,
            "usermetadata": res.usermetadata,
            "fieldmetadata": res.fieldmetadata,
            "security": res.security,
        }
        files_to_upload = []
        if res.data.conversations:
            data["conversations"] = dict(
                zip(
                    res.data.conversations.keys(),
                    [dict(v.value) for v in res.data.conversations.values()],
                )
            )
        if res.data.links:
            data["links"] = dict(
                zip(
                    res.data.links.keys(),
                    [dict(v.value) for v in res.data.links.values()],
                )
            )
        if res.data.texts:
            data["texts"] = dict(
                zip(
                    res.data.texts.keys(),
                    [dict(v.value) for v in res.data.texts.values()],
                )
            )
        if res.data.files:
            remote_files = {}
            for file_id, file in res.data.files.items():
                if file.value.external:
                    remote_files[file_id] = file.value
                else:
                    files_to_upload.append({"id": file_id, "data": file.value})
        destination_kb = await get_async_client(destination)
        failed = False
        try:
            uuid = await self.resource.create(ndb=destination_kb, **data)
        except exceptions.RateLimitError as e:
            failed = True
            delay = 5
            if e.try_after:
                delay = round(e.try_after - datetime.utcnow().timestamp())
            logger.warning(
                f"Backpressure error while copying resource, retrying in {delay} seconds"
            )
            await asyncio.sleep(delay)
            await self.copy(rid=rid, slug=slug, destination=destination, **kwargs)

        if not failed:
            with tempfile.TemporaryDirectory() as tmpdirname:
                for file_data in files_to_upload:
                    file_path = os.path.join(
                        tmpdirname, file_data["data"].file.filename
                    )
                    await self.resource.download_file(
                        rid=res.id, file_id=file_data["id"], output=file_path, **kwargs
                    )
                    await self.upload.file(path=file_path, ndb=destination_kb, rid=uuid)

    @kb
    async def copy_all(
        self, *, destination: str, page=0, filters: Optional[List[str]] = None, **kwargs
    ):
        ndb: AsyncNucliaDBClient = kwargs["ndb"]
        if filters is None:
            batch = await self.list(ndb=ndb, page=page)
            resources = batch.resources
            is_last = batch.pagination.last
        else:
            batch = await self.search.find(
                ndb=ndb, filters=filters, page=page, fields=["a/title"]
            )
            resources = batch.resources.values()
            is_last = batch.next_page
        for res in resources:
            try:
                logger.info(f"Copying resource {res.id}")
                await self.copy(rid=res.id, destination=destination, **kwargs)
            except exceptions.ConflictError:
                logger.info(f"Resource {res.id} already exists in destination KB")
        if not is_last:
            await self.copy_all(destination=destination, page=page + 1, **kwargs)



import random
import string

import pytest
from nuclia_models.accounts.backups import BackupCreate, BackupRestore

from nuclia import sdk
from nuclia.tests.fixtures import TESTING_ACCOUNT_SLUG, TESTING_KBID

ZONE = "europe-1"


def test_backup(testing_config):
    sdk.NucliaAccounts().default(TESTING_ACCOUNT_SLUG)

    # Create a backup
    backup = sdk.NucliaBackup().create(
        backup=BackupCreate(kb_id=TESTING_KBID),
        zone=ZONE,
    )
    assert backup.id is not None

    backups = sdk.NucliaBackup().list(zone=ZONE)
    backup_ids = [b.id for b in backups]
    assert backup.id in backup_ids

    # Restore the KB
    new_kb_slug = "backuptestsdk" + "".join(random.choices(string.ascii_lowercase, k=4))
    with pytest.raises(Exception) as exc_info:
        _ = sdk.NucliaBackup().restore(
            restore=BackupRestore(
                slug=new_kb_slug, title="Test SDK KB (can be deleted)"
            ),
            backup_id=backup.id,
            zone=ZONE,
        )

    # We don't want to wait till backup is ready to not slow down the test pipeline.
    # We already have a test for this on E2E repo
    expected_message = {
        "status": 409,
        "message": '{"detail":"Backup is still in progress. Please wait until it is completed."}',
    }
    assert exc_info.value.args[0] == expected_message

    # Delete the backup
    sdk.NucliaBackup().delete(id=backup.id, zone=ZONE)


def test_delete_all_backups(testing_config):
    sdk.NucliaAccounts().default(TESTING_ACCOUNT_SLUG)

    backups = sdk.NucliaBackup().list(zone=ZONE)
    for b in backups:
        # Delete backup
        sdk.NucliaBackup().delete(id=b.id, zone=ZONE)


@pytest.mark.asyncio
async def test_backup_async(testing_config):
    sdk.NucliaAccounts().default(TESTING_ACCOUNT_SLUG)

    # Create a backup
    backup = await sdk.AsyncNucliaBackup().create(
        backup=BackupCreate(kb_id=TESTING_KBID),
        zone=ZONE,
    )
    assert backup.id is not None

    backups = await sdk.AsyncNucliaBackup().list(zone=ZONE)
    backup_ids = [b.id for b in backups]
    assert backup.id in backup_ids

    # Restore the KB
    new_kb_slug = "backuptestsdk" + "".join(random.choices(string.ascii_lowercase, k=4))
    with pytest.raises(Exception) as exc_info:
        _ = await sdk.AsyncNucliaBackup().restore(
            restore=BackupRestore(
                slug=new_kb_slug, title="Test SDK KB (can be deleted)"
            ),
            backup_id=backup.id,
            zone=ZONE,
        )

    # We don't want to wait till backup is ready to not slow down the test pipeline.
    # We already have a test for this on E2E repo
    expected_message = {
        "status": 409,
        "message": '{"detail":"Backup is still in progress. Please wait until it is completed."}',
    }
    assert exc_info.value.args[0] == expected_message

    # Delete the backup
    await sdk.AsyncNucliaBackup().delete(id=backup.id, zone=ZONE)



from typing import Optional

from nuclia import get_regional_url
from nuclia.config import retrieve, retrieve_account
from nuclia.data import get_async_auth, get_auth
from nuclia.decorators import account, accounts, zone
from nuclia.sdk.auth import AsyncNucliaAuth, NucliaAuth

AGENTS_ENDPOINT = "/api/v1/account/{account}/kbs?mode=agent"
AGENT_ENDPOINT = "/api/v1/account/{account}/kb/{agent}"


class NucliaAgents:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @accounts
    def list(self, account: Optional[str] = None):
        if account is None:
            result = []
            accounts = (
                self._auth._config.accounts
                if self._auth._config.accounts is not None
                else []
            )
            for account_obj in accounts:
                if account_obj.slug is not None:
                    result.extend(self._auth.agents(account_obj.id))
            self._auth._config.agents = result
            self._auth._config.save()

            # List the Knowledge Boxes configured as Service Token
            result.extend(
                self._auth._config.agents_token
                if self._auth._config.agents_token is not None
                else []
            )
            return result
        else:
            matching_account = retrieve_account(
                self._auth._config.accounts or [], account
            )
            if not matching_account:
                raise ValueError("Account not found")
            return self._auth.agents(matching_account.id)

    @accounts
    @account
    @zone
    def get(
        self,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        **kwargs,
    ):
        zone = kwargs.get("zone")
        if not zone:
            raise ValueError("zone is required")
        if not id and not slug:
            raise ValueError("id or slug is required")
        if slug and not id:
            agents = self._auth.agents(kwargs["account_id"])
            agent_obj = retrieve(agents, slug)
            if not agent_obj:
                raise ValueError("Retrieval Agent not found")
            id = agent_obj.id
        path = get_regional_url(
            zone, AGENT_ENDPOINT.format(account=kwargs["account_id"], agent=id)
        )
        return self._auth._request("GET", path)

    def default(self, agent: str):
        agents = (
            self._auth._config.agents if self._auth._config.agents is not None else []
        )
        agent_obj = retrieve(agents, agent)
        if agent_obj is None:
            agents = (
                self._auth._config.agents_token
                if self._auth._config.agents_token is not None
                else []
            )
            agent_obj = retrieve(agents, agent)

        if agent_obj is None:
            raise KeyError("Retrieval Agent not found")
        self._auth._config.set_default_agent(agent_obj.id)
        self._auth._config.save()


class AsyncNucliaAgents:
    @property
    def _auth(self) -> AsyncNucliaAuth:
        auth = get_async_auth()
        return auth

    @accounts
    async def list(self, account: Optional[str] = None):
        if account is None:
            result = []
            accounts = (
                self._auth._config.accounts
                if self._auth._config.accounts is not None
                else []
            )
            for account_obj in accounts:
                if account_obj.slug is not None:
                    result.extend(await self._auth.agents(account_obj.id))
            self._auth._config.agents = result
            self._auth._config.save()

            # List the Knowledge Boxes configured as Service Token
            result.extend(
                self._auth._config.agents_token
                if self._auth._config.agents_token is not None
                else []
            )
            return result
        else:
            matching_account = retrieve_account(
                self._auth._config.accounts or [], account
            )
            if not matching_account:
                raise ValueError("Account not found")
            return await self._auth.agents(matching_account.id)

    @accounts
    @account
    @zone
    async def get(
        self,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        **kwargs,
    ):
        zone = kwargs.get("zone")
        if not zone:
            raise ValueError("zone is required")
        if not id and not slug:
            raise ValueError("id or slug is required")
        if slug and not id:
            agents = await self._auth.agents(kwargs["account_id"])
            agent_obj = retrieve(agents, slug)
            if not agent_obj:
                raise ValueError("Retrieval Agent not found")
            id = agent_obj.id
        path = get_regional_url(
            zone, AGENT_ENDPOINT.format(account=kwargs["account_id"], agent=id)
        )
        return await self._auth._request("GET", path)

    def default(self, agent: str):
        agents = (
            self._auth._config.agents if self._auth._config.agents is not None else []
        )
        agent_obj = retrieve(agents, agent)
        if agent_obj is None:
            agents = (
                self._auth._config.agents_token
                if self._auth._config.agents_token is not None
                else []
            )
            agent_obj = retrieve(agents, agent)

        if agent_obj is None:
            raise KeyError("Retrieval Agent not found")
        self._auth._config.set_default_agent(agent_obj.id)
        self._auth._config.save()



import pytest

from nuclia.sdk.auth import AsyncNucliaAuth, NucliaAuth
from nuclia.tests.fixtures import TESTING_KB, TESTING_KBID


def test_auth_user(testing_user: str):
    na = NucliaAuth()
    assert na._validate_user_token(testing_user)


def test_auth_kb(testing_kb: str):
    na = NucliaAuth()
    kbobj = na.validate_kb(TESTING_KB, testing_kb)
    assert kbobj
    assert kbobj.uuid
    assert kbobj.config
    assert kbobj.config.title


def test_auth_nua(testing_nua: str):
    na = NucliaAuth()
    client, account_type, account, region = na.validate_nua(testing_nua)
    assert client
    assert account_type
    assert account


def test_auth_pat(testing_config):
    na = NucliaAuth()
    token = na.create_personal_token(description="sdk test token", days=1, login=False)
    assert token
    tokens = na.list_personal_tokens()
    assert len([t.id for t in tokens if t.id == token.id]) == 1
    na.delete_personal_token(token_id=token.id)
    tokens = na.list_personal_tokens()
    assert len([t.id for t in tokens if t.id == token.id]) == 0


def test_auth_ephemeral_token(testing_config, testing_kb):
    na = NucliaAuth()
    token = na.create_ephemeral_token(kbid=TESTING_KBID)

    assert token
    assert token.token is not None

    token = na.create_ephemeral_token(kbid=TESTING_KBID, ttl=3600)

    assert token
    assert token.token is not None


def test_get_account_from_service_token(testing_kb: str):
    na = NucliaAuth()
    region = "europe-1"

    account_id, kb_id, role = na.get_account_from_service_token(region, testing_kb)

    assert account_id is not None
    assert kb_id is not None
    assert role is not None
    assert kb_id == TESTING_KBID


@pytest.mark.asyncio
async def test_get_account_from_service_token_async(testing_kb: str):
    na = AsyncNucliaAuth()
    region = "europe-1"

    account_id, kb_id, role = await na.get_account_from_service_token(
        region, testing_kb
    )

    assert account_id is not None
    assert kb_id is not None
    assert role is not None
    assert kb_id == TESTING_KBID


@pytest.mark.asyncio
async def test_auth_ephemeral_token_async(testing_config, testing_kb):
    na = AsyncNucliaAuth()
    token = await na.create_ephemeral_token(kbid=TESTING_KBID)

    assert token
    assert token.token is not None

    token = await na.create_ephemeral_token(kbid=TESTING_KBID, ttl=3600)

    assert token
    assert token.token is not None


@pytest.mark.asyncio
async def test_async_auth_user(testing_user: str):
    na = AsyncNucliaAuth()
    assert await na._validate_user_token(testing_user)


@pytest.mark.asyncio
async def test_async_auth_kb(testing_kb: str):
    na = AsyncNucliaAuth()
    kbobj = await na.validate_kb(TESTING_KB, testing_kb)
    assert kbobj
    assert kbobj.uuid
    assert kbobj.config
    assert kbobj.config.title


@pytest.mark.asyncio
async def test_async_auth_nua(testing_nua: str):
    na = AsyncNucliaAuth()
    client, account_type, account, region = await na.validate_nua(testing_nua)
    assert client
    assert account_type
    assert account


@pytest.mark.asyncio
async def test_async_auth_kb_url(testing_kb: str):
    na = AsyncNucliaAuth()
    kb_uuid = await na.kb(TESTING_KB, testing_kb)
    assert kb_uuid == TESTING_KBID


@pytest.mark.asyncio
async def test_async_auth_nua_token(testing_nua: str):
    na = AsyncNucliaAuth()
    client_id = await na.nua(testing_nua)
    assert client_id is not None


@pytest.mark.asyncio
async def test_async_auth_accounts(testing_config):
    na = AsyncNucliaAuth()
    accounts = await na.accounts()
    assert len(accounts) > 0
    assert accounts[0].id
    assert accounts[0].slug


@pytest.mark.asyncio
async def test_async_auth_zones(testing_config):
    na = AsyncNucliaAuth()
    zones = await na.zones()
    assert len(zones) > 0
    assert zones[0].slug


@pytest.mark.asyncio
async def test_async_auth_kbs(testing_config):
    na = AsyncNucliaAuth()
    accounts = await na.accounts()
    assert len(accounts) > 0
    kbs = await na.kbs(accounts[0].id)
    # Should return a list (may be empty depending on account)
    assert isinstance(kbs, list)


@pytest.mark.asyncio
async def test_async_auth_agents(testing_config):
    na = AsyncNucliaAuth()
    accounts = await na.accounts()
    assert len(accounts) > 0
    agents = await na.agents(accounts[0].id)
    # Should return a list (may be empty depending on account)
    assert isinstance(agents, list)


def test_auth_agent_url(testing_agent: str):
    from nuclia.tests.fixtures import TESTING_AGENT, TESTING_AGENT_ID

    na = NucliaAuth()
    agent_uuid = na.agent(TESTING_AGENT, testing_agent)
    assert agent_uuid == TESTING_AGENT_ID


@pytest.mark.asyncio
async def test_async_auth_agent_url(testing_agent: str):
    from nuclia.tests.fixtures import TESTING_AGENT, TESTING_AGENT_ID

    na = AsyncNucliaAuth()
    agent_uuid = await na.agent(TESTING_AGENT, testing_agent)
    assert agent_uuid == TESTING_AGENT_ID



from nuclia.sdk.kb import AsyncNucliaKB, NucliaKB


def test_split_strategies(testing_config):
    nkb = NucliaKB()
    # preventive clean up
    for id in nkb.split_strategies.list().keys():
        nkb.split_strategies.delete(id=id)

    # tests
    nkb.split_strategies.add(config={"name": "strategy1", "max_paragraph": 400})
    all = nkb.split_strategies.list()
    assert len(all.keys()) == 1
    nkb.split_strategies.delete(id=list(all.keys())[0])
    all = nkb.split_strategies.list()
    assert len(all.keys()) == 0


async def test_split_strategies_async(testing_config):
    nkb = AsyncNucliaKB()
    # preventive clean up
    strategies_ids = await nkb.split_strategies.list()
    for id in strategies_ids.keys():
        await nkb.split_strategies.delete(id=id)

    # tests
    await nkb.split_strategies.add(config={"name": "strategy1", "max_paragraph": 400})
    all = await nkb.split_strategies.list()
    assert len(all.keys()) == 1
    await nkb.split_strategies.delete(id=list(all.keys())[0])
    all = await nkb.split_strategies.list()
    assert len(all.keys()) == 0



import asyncio
from typing import Type, Union
from uuid import uuid4

import pytest

from nuclia.sdk.agent_sessions import AsyncNucliaAgentSessions, NucliaAgentSessions
from nuclia.tests.utils import maybe_await


@pytest.mark.parametrize(
    "agent_sessions_klass",
    [NucliaAgentSessions, AsyncNucliaAgentSessions],
)
async def test_sessions(
    testing_config,
    agent_sessions_klass: Type[Union[NucliaAgentSessions, AsyncNucliaAgentSessions]],
):
    n_agent_sess = agent_sessions_klass()

    session_name = f"nuclia-py-test-{uuid4().hex}"
    # Create session
    uuid = await maybe_await(n_agent_sess.new(session_name))

    # Get session
    session = await maybe_await(n_agent_sess.get(uuid))
    assert session.title == session_name

    # List sessions, scan all pages since the endpoint does not support sorting by creation date
    last = False
    found = False
    p = 0
    while not last:
        sessions = await maybe_await(n_agent_sess.list(page_size=100, page=p))
        last = sessions.pagination.last
        p += 1
        found = any(s.id == uuid for s in sessions.resources)
        if found:
            break
    assert found

    # Delete session
    await maybe_await(n_agent_sess.delete(uuid))
    # Wait a bit for deletion to propagate
    await asyncio.sleep(1)
    with pytest.raises(Exception):
        await maybe_await(n_agent_sess.get(uuid))



import os
from typing import List, Optional
from urllib.parse import urlparse

BASE_DOMAIN = os.environ.get("BASE_NUCLIA_DOMAIN", "rag.progress.cloud")
BASE = f"https://{BASE_DOMAIN}"
REGIONAL = "https://{region}." + BASE_DOMAIN
CLOUD_ID = BASE.split("/")[-1]


def get_global_url(path: str):
    return BASE + path


def get_regional_url(region: str, path: str):
    return REGIONAL.format(region=region) + path


def get_list_parameter(param: Optional[List[str]]) -> List[str]:
    param = list(param or [])
    if len(param) > 0 and all([len(el) == 1 for el in param]):
        # Python Fire converts single string to list of chars
        param = ["".join(param)]
    return param


def is_nuclia_hosted(url: str):
    return BASE_DOMAIN in urlparse(url).netloc


# HACK used to debug in CI. Remove once the issue is solved

_DEBUG_HTTPX_REQUESTS = os.environ.get("DEBUG_HTTPX_REQUESTS", "false").lower() in {
    "true",
    "1",
}

if _DEBUG_HTTPX_REQUESTS:
    import httpx
    from wrapt import wrap_function_wrapper  # type: ignore

    def sync_wrapper(wrapped, instance: httpx.HTTPTransport, args, kwargs):
        print(f"[httpx::handle_request] {args[0].method} {args[0].url}")
        return wrapped(*args, **kwargs)

    async def async_wrapper(wrapped, instance: httpx.AsyncHTTPTransport, args, kwargs):
        print(f"[httpx::handle_async_request] {args[0].method} {args[0].url}")
        return await wrapped(*args, **kwargs)

    wrap_function_wrapper("httpx", "HTTPTransport.handle_request", sync_wrapper)
    wrap_function_wrapper(
        "httpx", "AsyncHTTPTransport.handle_async_request", async_wrapper
    )



from .accounts import NucliaAccounts
from .agent import NucliaAgent
from .auth import AsyncNucliaAuth, NucliaAuth
from .backup import AsyncNucliaBackup, NucliaBackup
from .export_import import AsyncNucliaExports, NucliaExports, NucliaImports
from .kb import AsyncNucliaKB, NucliaKB
from .kbs import AsyncNucliaKBS, NucliaKBS
from .nua import NucliaNUA
from .nucliadb import NucliaDB
from .predict import AsyncNucliaPredict, NucliaPredict
from .resource import AsyncNucliaResource, NucliaResource
from .search import AsyncNucliaSearch, NucliaSearch
from .task import NucliaTask
from .upload import AsyncNucliaUpload, NucliaUpload
from .zones import NucliaZones

__all__ = [
    "NucliaAccounts",
    "NucliaAgent",
    "AsyncNucliaAuth",
    "NucliaAuth",
    "AsyncNucliaExports",
    "NucliaExports",
    "NucliaImports",
    "AsyncNucliaKB",
    "NucliaKB",
    "NucliaKBS",
    "AsyncNucliaKBS",
    "NucliaBackup",
    "AsyncNucliaBackup",
    "NucliaNUA",
    "NucliaDB",
    "AsyncNucliaPredict",
    "NucliaPredict",
    "AsyncNucliaResource",
    "NucliaResource",
    "AsyncNucliaSearch",
    "NucliaSearch",
    "NucliaTask",
    "AsyncNucliaUpload",
    "NucliaUpload",
    "NucliaZones",
]



import pytest

from nuclia.sdk.kb import NucliaKB


def test_labels(testing_config):
    nkb = NucliaKB()
    nkb.set_labelset(labelset="labelset1")
    nkb.add_label(labelset="labelset1", label="label1")
    nkb.add_labels(labelset="labelset1", labels=["label1", "label2"])
    with pytest.raises(ValueError):
        nkb.add_labels(labelset="whatever", labels=["label1", "label2"])
    nkb.del_labelset(labelset="labelset2")
    labelset = nkb.get_labelset(labelset="labelset1")
    assert len(labelset.labels) == 2

    nkb.del_label(labelset="labelset1", label="label2")

    labelset = nkb.get_labelset(labelset="labelset1")
    assert len(labelset.labels) == 1
    assert labelset.labels[0].title == "label1"

    labelsets = nkb.list_labelsets()
    assert len(labelsets.labelsets) == 1
    assert "labelset1" in labelsets.labelsets
    nkb.del_labelset(labelset="labelset1")
    labelsets = nkb.list_labelsets()
    assert len(labelsets.labelsets) == 0



from datetime import datetime
from enum import Enum, IntEnum
from typing import Any, Dict, List, Literal, Optional, Union, cast

import pydantic
from nuclia_models.common.consumption import Consumption
from pydantic import BaseModel, Field, RootModel, model_validator
from typing_extensions import Annotated, Self


class GenerativeOption(BaseModel):
    name: str
    value: str
    user_prompt: Optional[str] = None
    user_key: Optional[str] = None


class Option(BaseModel):
    name: str
    value: str


class ConfigSchemaOptions(BaseModel):
    options: Union[List[GenerativeOption], List[Option]]
    default: Optional[str] = None
    create: bool = False
    update: bool = False
    multiple: bool = False


class ConfigSchemaElements(BaseModel):
    schemas: Dict[str, Any]
    create: bool = False
    update: bool = False


class ConfigSchema(BaseModel):
    semantic_model: ConfigSchemaOptions
    anonymization_model: ConfigSchemaOptions
    visual_labeling: Optional[ConfigSchemaOptions] = None
    generative_model: Optional[ConfigSchemaOptions] = None
    ner_model: Optional[ConfigSchemaOptions] = None
    relation_model: Optional[ConfigSchemaOptions] = None
    summary_model: Optional[ConfigSchemaOptions] = None
    summary: Optional[ConfigSchemaOptions] = None
    user_keys: Optional[ConfigSchemaElements] = None
    user_prompts: Optional[ConfigSchemaElements] = None
    summary_prompt: Optional[ConfigSchemaElements] = None
    prefer_markdown_generative_response: Optional[ConfigSchemaElements] = None


class Sentence(BaseModel):
    data: List[float]
    time: float
    consumption: Optional[Consumption] = None


class Author(str, Enum):
    NUCLIA = "NUCLIA"
    USER = "USER"


class Message(BaseModel):
    author: Author
    text: str


class UserPrompt(BaseModel):
    prompt: str


class Image(BaseModel):
    content_type: str
    b64encoded: str


class Tool(BaseModel):
    name: str
    description: str
    parameters: Dict[str, Any] = Field(
        default_factory=dict, description="Schema of the tool"
    )


class Reasoning(BaseModel):
    display: bool = Field(
        default=True,
        description="Whether to display the reasoning steps in the response.",
    )
    effort: Literal["low", "medium", "high"] = Field(
        default="medium",
        description=(
            "Level of reasoning effort. Used by OpenAI models to control the depth of reasoning. "
            "This parameter will be automatically mapped to budget_tokens "
            "if the chosen model does not support effort."
        ),
    )
    budget_tokens: int = Field(
        default=15_000,
        description=(
            "Token budget for reasoning. Used by Anthropic or Google models to limit the number of "
            "tokens used for reasoning. This parameter will be automatically mapped to effort "
            "if the chosen model does not support budget_tokens."
        ),
    )

    @model_validator(mode="before")
    @classmethod
    def set_budget_tokens_or_effort(cls, data: Any) -> Any:
        if isinstance(data, dict):
            if "budget_tokens" not in data and "effort" not in data:
                return data

            effort_map = {"low": 7500, "medium": 15_000, "high": 30_000}
            if "budget_tokens" not in data:
                if data["effort"] in effort_map:
                    data["budget_tokens"] = effort_map[data["effort"]]
                else:
                    raise ValueError(
                        f"Invalid effort value: {data['effort']}. "
                        f"Valid values are: {', '.join(effort_map.keys())}."
                    )
            if "effort" not in data:
                budget_tokens = data["budget_tokens"]
                if not isinstance(budget_tokens, int):
                    raise ValueError(
                        f"Invalid budget_tokens value: {budget_tokens}. "
                        "It must be an integer."
                    )
                if budget_tokens <= effort_map["low"]:
                    data["effort"] = "low"
                elif budget_tokens <= effort_map["medium"]:
                    data["effort"] = "medium"
                else:
                    data["effort"] = "high"

        return data


class CitationsType(str, Enum):
    NONE = "none"
    DEFAULT = "default"
    LLM_FOOTNOTES = "llm_footnotes"


class ToolChoiceAuto(BaseModel):
    type: Literal["auto"] = "auto"


class ToolChoiceNone(BaseModel):
    type: Literal["none"] = "none"


class ToolChoiceRequired(BaseModel):
    type: Literal["required"] = "required"


class ToolChoiceForced(BaseModel):
    type: Literal["forced"] = "forced"
    name: str


class ChatModel(BaseModel):
    question: str
    retrieval: bool = True
    user_id: str
    system: Optional[str] = None
    chat_history: List[Message] = []
    context: List[Message] = []
    query_context: Union[List[str], Dict[str, str]] = {}
    query_context_order: Dict[str, int] = {}
    truncate: Optional[bool] = True
    user_prompt: Optional[UserPrompt] = None
    citations: Union[bool, None, CitationsType] = Field(
        default=None,
        description="Whether to include citations in the response. "
        "If set to None or False, no citations will be computed. "
        "If set to True or 'default', citations will be computed after answer generation and send as a separate `CitationsGenerativeResponse` chunk"
        "If set to 'llm_footnotes', citations will be included in the LLM's response as markdown-styled footnotes. A `FootnoteCitationsGenerativeResponse` chunk will also be sent to map footnote ids to context keys in the `query_context`.",
    )
    citation_threshold: Optional[float] = Field(
        default=None,
        description="If citations is set to True, this will be the similarity threshold. Value between 0 and 1, lower values will produce more citations. If not set, it will be set to the optimized threshold found by Nuclia.",
        ge=0.0,
        le=1.0,
    )
    generative_model: Optional[str] = None
    max_tokens: Optional[int] = None
    query_context_images: Union[
        List[Image], Dict[str, Image]
    ] = {}  # base64.b64encode(image_file.read()).decode('utf-8')
    prefer_markdown: Optional[bool] = None
    json_schema: Optional[Dict[str, Any]] = None
    format_prompt: bool = True
    rerank_context: bool = Field(
        default=False,
        description="Whether to reorder the query context based on a reranker. This option will also make it so the first response will contain the scores given for each context piece.",
    )
    tools: List[Tool] = Field(
        default_factory=list, description="List of tools to choose"
    )
    tool_choice: Union[
        ToolChoiceAuto, ToolChoiceNone, ToolChoiceRequired, ToolChoiceForced
    ] = Field(
        default=ToolChoiceRequired(),
        discriminator="type",
        description=(
            "Tool choice strategy. "
            "`auto`: The model decides whether to use a tool or not based on the prompt and available tools. "
            "`required` (default): A tool must be used."
            "`none`: Disables tool usage even if tools are provided."
            "`forced`: Forces the use of a specific tool provided in `name`."
            "Important: Not all model providers support all tool choice strategies, its a best effort feature."
        ),
    )
    reasoning: Union[Reasoning, bool] = Field(
        default=False,
        description=(
            "Reasoning options for the generative model. "
            "Set to True to enable default reasoning, False to disable, or provide a Reasoning object for custom options."
        ),
    )

    @model_validator(mode="after")
    def validate_model(self) -> Self:
        if self.prefer_markdown is True and self.json_schema is not None:
            raise ValueError("Can not setup markdown and JSON Schema at the same time")
        if self.citations is True and self.json_schema is not None:
            raise ValueError("Can not setup citations and JSON Schema at the same time")
        if self.citations is True and len(self.tools) > 0:
            raise ValueError("Can not setup citations and Tools at the same time")
        if len(self.tools) > 0 and self.json_schema is not None:
            raise ValueError("Can not setup Tools and JSON Schema at the same time")
        return self

    @model_validator(mode="after")
    def check_reasoning_budget_vs_max_tokens(self) -> "ChatModel":
        if isinstance(self.reasoning, Reasoning) and self.max_tokens is not None:
            if self.reasoning.budget_tokens >= self.max_tokens:
                raise ValueError(
                    f"`budget_tokens` ({self.reasoning.budget_tokens}) cannot be greater or equal than `max_tokens` ({self.max_tokens})."
                )
        return self


class Token(BaseModel):
    text: str
    ner: str
    start: int
    end: int


class Tokens(BaseModel):
    tokens: List[Token]
    time: float
    consumption: Optional[Consumption] = None


class SummarizeResource(BaseModel):
    fields: Dict[str, str]


class SummarizeModel(BaseModel):
    resources: Dict[str, SummarizeResource]


class SummarizedResource(BaseModel):
    summary: str
    tokens: int


class SummarizedModel(BaseModel):
    resources: Dict[str, SummarizedResource]
    summary: str = ""
    consumption: Optional[Consumption] = None


class RephraseModel(RootModel[str]):
    pass


class WebhookConfig(BaseModel):
    uri: Optional[str] = None
    headers: Dict[str, str] = {}


# Pull models


class PullStatus(str, Enum):
    OK = "ok"
    ERROR = "error"
    EMPTY = "empty"


class PullResponse(BaseModel):
    status: PullStatus
    payload: Optional[bytes]
    msgid: Optional[str]


# Push models


class ProcessingQueueType(str, Enum):
    SHARED = "shared"
    PRIVATE = "private"


class Format(int, Enum):
    PLAIN = 0
    HTML = 1
    MARKDOWN = 2
    RST = 3
    JSON = 4


class ProcessingMessageContent(BaseModel):
    text: Optional[str] = None
    format: Format
    attachments: Optional[List[str]] = None


class ProcessingMessage(BaseModel):
    timestamp: Optional[datetime] = None
    who: Optional[str] = None
    to: List[str] = []
    content: ProcessingMessageContent
    ident: str


class ProcessingConversation(BaseModel):
    messages: List[ProcessingMessage]


class LayoutFormat(int, Enum):
    FLAPSv1 = 0


class TypeBlock(int, Enum):
    TITLE = 0
    DESCRIPTION = 1
    RICHTEXT = 2
    TEXT = 3
    ATTACHMENTS = 4
    COMMENTS = 5
    CLASSIFICATIONS = 6


class ProcessingBlock(BaseModel):
    x: int
    y: int
    cols: int
    rows: int
    type: TypeBlock
    ident: str
    payload: str
    file: Optional[str] = None


class ProcessingLayoutDiff(BaseModel):
    format: LayoutFormat
    blocks: Dict[str, ProcessingBlock]


class LinkUpload(BaseModel):
    link: str
    headers: Dict[str, str]
    cookies: Dict[str, str]
    localstorage: Dict[str, str]


class Text(BaseModel):
    body: str
    format: Format


class Source(int, Enum):
    HTTP = 0
    INGEST = 1


def validate_uuid(value, handler, info):
    if not value:
        raise ValueError(f"Invalid uuid: '{value}'. Uuid must be a non-empty string.")
    try:
        return handler(value)
    except pydantic.ValidationError as e:
        if any(x["type"] == "string_pattern_mismatch" for x in e.errors()):
            raise ValueError(
                f"Invalid slug: '{value}'. Slug must be a string with only "
                "letters, numbers, underscores, colons and dashes."
            )
        else:
            raise e


RestrictedIDString = Annotated[
    str,
    pydantic.StringConstraints(pattern=r"^[a-z0-9_-]+$"),
    pydantic.WrapValidator(validate_uuid),
]


class PushProcessingOptions(BaseModel):
    # Enable ML processing
    ml_text: bool = True


class PushPayload(BaseModel):
    kbid: Optional[RestrictedIDString] = cast(RestrictedIDString, "default")
    uuid: Optional[RestrictedIDString] = None

    slug: Optional[str] = None
    source: Optional[Source] = None
    userid: Optional[str] = None

    # Generic Field
    genericfield: Dict[str, Text] = {}

    # New File
    filefield: Dict[str, str] = {}

    # New Link
    linkfield: Dict[str, LinkUpload] = {}

    # Diff on Text Field
    textfield: Dict[str, Text] = {}

    # Diff on a Layout Field
    layoutfield: Dict[str, ProcessingLayoutDiff] = {}

    # New conversations to process
    conversationfield: Dict[str, ProcessingConversation] = {}

    # List of available processing options (with default values)
    processing_options: Optional[PushProcessingOptions] = Field(
        default_factory=PushProcessingOptions
    )


class PublicPushPayload(PushPayload):
    # If provided, will override the webhookconfig on the nua-token (if any)
    webhook_config: Optional[WebhookConfig] = None


class PublicPushResponse(BaseModel):
    seqid: Optional[int] = None
    account_seq: Optional[int] = None
    queue: Optional[ProcessingQueueType] = None
    # On Public api, uuid may be generated by the proxy, so we need to return it to the user
    uuid: Optional[str] = None


class ProcessingStatusInfo(BaseModel):
    last_delivered_seqid: Optional[int] = (
        None  # When none, means we already don't have information about this queue
    )


class ProcessingStatus(BaseModel):
    shared: ProcessingStatusInfo
    account: Optional[ProcessingStatusInfo]


class ProcessRequestStatus(BaseModel):
    processing_id: str
    resource_id: Optional[str]
    kbid: Optional[str]
    title: Optional[str]
    labels: list[str]
    completed: bool
    scheduled: bool
    timestamp: datetime
    completed_at: Optional[datetime]
    scheduled_at: Optional[datetime]
    failed: bool = False
    retries: int = 0
    schedule_eta: float = 0.0
    schedule_order: int = 0
    response: Optional[str] = None


class ProcessRequestStatusResults(BaseModel):
    results: list[ProcessRequestStatus]
    cursor: Optional[str]


class PushResponseV2(BaseModel):
    processing_id: str


class Summary(str, Enum):
    EXTENDED = "extended"
    SIMPLE = "simple"


class SummaryPrompt(BaseModel):
    prompt: str


class OpenAIKey(BaseModel):
    key: str
    org: str


class AzureOpenAIKey(BaseModel):
    key: str
    url: str
    deployment: str
    model: str


class PalmKey(BaseModel):
    credentials: str
    location: str


class CohereKey(BaseModel):
    key: str


class AnthropicKey(BaseModel):
    key: str


class OpenAIUserPrompt(BaseModel):
    system: str
    prompt: str


class AzureUserPrompt(BaseModel):
    system: str
    prompt: str


class PalmUserPrompt(BaseModel):
    prompt: str


class CohereUserPrompt(BaseModel):
    prompt: str


class AnthropicUserPrompt(BaseModel):
    prompt: str


class TextGenerationUserPrompt(BaseModel):
    prompt: str


class UserPrompts(BaseModel):
    openai: Optional[OpenAIUserPrompt] = None
    azure_openai: Optional[AzureUserPrompt] = None
    palm: Optional[PalmUserPrompt] = None
    cohere: Optional[CohereUserPrompt] = None
    anthropic: Optional[AnthropicUserPrompt] = None
    text_generation: Optional[TextGenerationUserPrompt] = None


class UserLearningKeys(BaseModel):
    openai: Optional[OpenAIKey] = None
    azure_openai: Optional[AzureOpenAIKey] = None
    palm: Optional[PalmKey] = None
    cohere: Optional[CohereKey] = None
    anthropic: Optional[AnthropicKey] = None


class LearningConfigurationUpdate(BaseModel):
    # Validate its enum or string
    anonymization_model: Optional[str] = None
    visual_labeling: Optional[str] = None
    generative_model: Optional[str] = None
    ner_model: Optional[str] = None
    relation_model: Optional[str] = None
    user_keys: Optional[UserLearningKeys] = None
    user_prompts: Optional[UserPrompts] = None
    summary: Optional[Summary] = Summary.SIMPLE
    summary_model: Optional[str] = None
    summary_prompt: Optional[SummaryPrompt] = None
    resource_labelers_models: Optional[List[str]] = None
    paragraph_labelers_models: Optional[List[str]] = None
    intent_models: Optional[List[str]] = None


class LearningConfigurationCreation(LearningConfigurationUpdate):
    semantic_model: Optional[str] = None


class Empty(BaseModel):
    pass


class ChatResponse(BaseModel):
    answer: str

    @classmethod
    def model_validate(
        cls,
        obj: bytes,
        **kwargs,
    ) -> "ChatResponse":
        return ChatResponse(answer=obj.decode())


class SimilarityFunction(IntEnum):
    # Keep this in sync with SimilarityFunction in config.proto
    # It's an IntEnum to match the protobuf definition
    DOT = 0
    COSINE = 1


class SemanticConfig(BaseModel):
    # Keep this in sync with SemanticConfig in config.proto
    similarity: SimilarityFunction
    size: int
    threshold: float
    max_tokens: Optional[int] = None
    matryoshka_dims: list[int] = []
    external: bool = False


class StoredLearningConfiguration(BaseModel):
    resource_labelers_models: Optional[List[str]] = None
    paragraph_labelers_models: Optional[List[str]] = None
    intent_models: Optional[List[str]] = None

    default_semantic_model: Optional[str] = None
    semantic_models: Optional[list[str]] = Field(default_factory=list[str])
    semantic_model_configs: dict[str, SemanticConfig] = {}
    semantic_model: str
    anonymization_model: str
    generative_model: str
    ner_model: str
    relation_model: str
    visual_labeling: Optional[str] = None

    user_keys: Optional[UserLearningKeys] = None
    user_prompts: Optional[UserPrompts] = None

    semantic_vector_similarity: str

    semantic_vector_size: int

    semantic_threshold: float

    summary: str
    summary_model: str
    summary_prompt: Optional[SummaryPrompt] = None


class SentenceSearch(BaseModel):
    data: List[float] = []
    time: float
    consumption: Optional[Consumption] = None


class Ner(BaseModel):
    text: str
    ner: str
    start: int
    end: int


class TokenSearch(BaseModel):
    tokens: List[Ner] = []
    time: float
    consumption: Optional[Consumption] = None


class QueryInfo(BaseModel):
    language: str
    stop_words: List[str]
    semantic_threshold: float
    visual_llm: bool
    max_context: int
    entities: Optional[TokenSearch]
    sentence: Optional[SentenceSearch]


class RerankModel(BaseModel):
    question: str
    user_id: str
    context: dict[str, str] = {}


class RerankResponse(BaseModel):
    context_scores: dict[str, float] = Field(
        description="Scores for each context given by the reranker"
    )
    consumption: Optional[Consumption] = None



import pytest
from nuclia_models.events.activity_logs import (
    ActivityLogsQuery,
    DownloadActivityLogsQuery,
    DownloadFormat,
    EventType,
    Pagination,
)

from nuclia.sdk.kb import AsyncNucliaKB, NucliaKB
from nuclia.tests.fixtures import IS_PROD


def test_activity_logs_query(testing_config):
    if not IS_PROD:
        assert True
        return
    query = ActivityLogsQuery(
        year_month="2024-10",
        show=["id", "date", "client_type", "total_duration"],
        filters={},
        pagination=Pagination(limit=10),
    )
    nkb = NucliaKB()
    output = nkb.logs.query(type=EventType.ASK, query=query)
    assert len(output.data) == 10
    assert output.has_more


def test_activity_logs_download(testing_config):
    if not IS_PROD:
        assert True
        return
    query = DownloadActivityLogsQuery(
        year_month="2024-10",
        show=["id", "date", "client_type", "total_duration"],
        filters={},
    )
    nkb = NucliaKB()
    output = nkb.logs.download(
        type=EventType.ASK, query=query, download_format=DownloadFormat.NDJSON
    )
    assert output.request_id
    assert output.download_url is None


@pytest.mark.asyncio
async def test_activity_logs_query_async(testing_config):
    if not IS_PROD:
        assert True
        return
    query = ActivityLogsQuery(
        year_month="2024-10",
        show=["id", "date", "client_type", "total_duration"],
        filters={},
        pagination=Pagination(limit=10),
    )
    nkb = AsyncNucliaKB()
    output = await nkb.logs.query(type=EventType.ASK, query=query)
    assert len(output.data) == 10
    assert output.has_more


@pytest.mark.asyncio
async def test_activity_logs_download_async(testing_config):
    if not IS_PROD:
        assert True
        return
    query = DownloadActivityLogsQuery(
        year_month="2024-10",
        show=["id", "date", "client_type", "total_duration"],
        filters={},
    )
    nkb = AsyncNucliaKB()
    output = await nkb.logs.download(
        type=EventType.ASK, query=query, download_format=DownloadFormat.NDJSON
    )
    assert output.request_id
    assert output.download_url is None


def test_activity_logs_ask_query(testing_config):
    if not IS_PROD:
        assert True
        return
    query = ActivityLogsQuery(
        year_month="2024-10",
        show=["id", "date", "client_type", "total_duration"],
        filters={},
        pagination=Pagination(limit=10),
    )
    nkb = NucliaKB()
    output = nkb.logs.query(type=EventType.ASK, query=query)
    assert len(output.data) == 10
    assert output.has_more


@pytest.mark.asyncio
async def test_activity_logs_ask_download_async(testing_config):
    if not IS_PROD:
        assert True
        return
    query = DownloadActivityLogsQuery(
        year_month="2024-10",
        show=["id", "date", "client_type", "total_duration"],
        filters={},
    )
    nkb = AsyncNucliaKB()
    output = await nkb.logs.download(
        type=EventType.ASK, query=query, download_format=DownloadFormat.NDJSON
    )
    assert output.request_id
    assert output.download_url is None



from nuclia.config import extract_region


def test_extract_region(monkeypatch):
    assert extract_region("https://europe-1.rag.progress.cloud/api/v1") == "europe-1"
    assert extract_region("https://europe-1.nuclia.cloud/api/v1") == "europe-1"
    assert extract_region("https://europe-1.stashify.cloud/api/v1") == "europe-1"
    assert (
        extract_region("https://europe-1.gcp-global-dev-1.nuclia.io/api/v1")
        == "europe-1"
    )
    assert extract_region("https://rag.progress.cloud/api/v1") is None
    assert extract_region("https://nuclia.cloud/api/v1") is None
    assert extract_region("https://stashify.cloud/api/v1") is None
    assert extract_region("https://gcp-global-dev-1.nuclia.io/api/v1") is None

    assert extract_region("https://europe-1.example.com/api/v1") == "europe-1"
    assert extract_region("https://example.com/api/v1") == "example"

    assert extract_region("https://localhost:8000/api/v1") is None

    assert extract_region("not a url") is None
    # Custom base domain
    monkeypatch.setattr("nuclia.config.CLOUD_ID", "example.com")
    assert extract_region("https://europe-1.example.com/api/v1") == "europe-1"
    assert extract_region("https://example.com/api/v1") is None



import pytest

from nuclia.sdk.agents import AsyncNucliaAgents, NucliaAgents
from nuclia.tests.fixtures import IS_PROD, TESTING_ACCOUNT_SLUG, TESTING_AGENT_ID


def test_list_agents(testing_config):
    agents = NucliaAgents()
    all = agents.list()
    if not IS_PROD and TESTING_AGENT_ID:
        assert TESTING_AGENT_ID in [agent.id for agent in all]


@pytest.mark.asyncio
async def test_async_list_agents(testing_config):
    agents = AsyncNucliaAgents()
    all = await agents.list()
    if not IS_PROD and TESTING_AGENT_ID:
        assert TESTING_AGENT_ID in [agent.id for agent in all]


def test_get_agent(testing_config):
    if not TESTING_AGENT_ID:
        # Skip if no agent configured
        assert True
        return
    agents = NucliaAgents()
    agent = agents.get(
        account=TESTING_ACCOUNT_SLUG, id=TESTING_AGENT_ID, zone="europe-1"
    )
    assert agent is not None
    assert agent["id"] == TESTING_AGENT_ID


@pytest.mark.asyncio
async def test_async_get_agent(testing_config):
    if not TESTING_AGENT_ID:
        # Skip if no agent configured
        assert True
        return
    agents = AsyncNucliaAgents()
    agent = await agents.get(
        account=TESTING_ACCOUNT_SLUG, id=TESTING_AGENT_ID, zone="europe-1"
    )
    assert agent is not None
    assert agent["id"] == TESTING_AGENT_ID



from typing import Dict, Optional

from nuclia import get_regional_url
from nuclia.config import retrieve, retrieve_account
from nuclia.data import get_async_auth, get_auth
from nuclia.decorators import account, accounts, zone
from nuclia.sdk.auth import AsyncNucliaAuth, NucliaAuth

KBS_ENDPOINT = "/api/v1/account/{account}/kbs"
KB_ENDPOINT = "/api/v1/account/{account}/kb/{kb}"


class NucliaKBS:
    @property
    def _auth(self) -> NucliaAuth:
        auth = get_auth()
        return auth

    @accounts
    def list(self, account: Optional[str] = None):
        if account is None:
            result = []
            accounts = (
                self._auth._config.accounts
                if self._auth._config.accounts is not None
                else []
            )
            for account_obj in accounts:
                if account_obj.slug is not None:
                    result.extend(self._auth.kbs(account_obj.id))
            self._auth._config.kbs = result
            self._auth._config.save()

            # List the Knowledge Boxes configured as Service Token
            result.extend(
                self._auth._config.kbs_token
                if self._auth._config.kbs_token is not None
                else []
            )
            return result
        else:
            matching_account = retrieve_account(
                self._auth._config.accounts or [], account
            )
            if not matching_account:
                raise ValueError("Account not found")
            return self._auth.kbs(matching_account.id)

    @accounts
    @account
    @zone
    def add(
        self,
        slug: str,
        anonymization: Optional[str] = None,
        description: Optional[str] = None,
        learning_configuration: Optional[Dict[str, str]] = None,
        sentence_embedder: Optional[str] = None,
        title: Optional[str] = None,
        zone: Optional[str] = "europe-1",
        **kwargs,
    ):
        if not slug:
            raise ValueError("slug is required.")
        if not zone:
            raise ValueError("zone is required")
        path = get_regional_url(zone, KBS_ENDPOINT.format(account=kwargs["account_id"]))
        data = {
            "slug": slug,
            "anonymization": anonymization,
            "description": description,
            "learning_configuration": learning_configuration,
            "sentence_embedder": sentence_embedder,
            "title": title or slug,
            "zone": zone,
        }
        self._auth._request("POST", path, data)
        return self.get(slug, account_id=kwargs["account_id"], zone=zone)

    @accounts
    @account
    @zone
    def get(
        self,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        **kwargs,
    ):
        zone = kwargs.get("zone")
        if not zone:
            raise ValueError("zone is required")
        if not id and not slug:
            raise ValueError("id or slug is required")
        if slug and not id:
            kbs = self._auth.kbs(kwargs["account_id"])
            kb_obj = retrieve(kbs, slug)
            if not kb_obj:
                raise ValueError("Knowledge Box not found")
            id = kb_obj.id
        path = get_regional_url(
            zone, KB_ENDPOINT.format(account=kwargs["account_id"], kb=id)
        )
        return self._auth._request("GET", path)

    @accounts
    @account
    @zone
    def delete(
        self,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        **kwargs,
    ):
        zone = kwargs.get("zone")
        if not zone:
            raise ValueError("zone is required")
        if not id and not slug:
            raise ValueError("id or slug is required")
        if slug and not id:
            kbs = self._auth.kbs(kwargs["account_id"])
            kb_obj = retrieve(kbs, slug)
            if not kb_obj:
                raise ValueError("Knowledge Box not found")
            id = kb_obj.id
        path = get_regional_url(
            zone, KB_ENDPOINT.format(account=kwargs["account_id"], kb=id)
        )
        return self._auth._request("DELETE", path)

    def default(self, kb: str):
        kbs = self._auth._config.kbs if self._auth._config.kbs is not None else []
        kb_obj = retrieve(kbs, kb)
        if kb_obj is None:
            kbs = (
                self._auth._config.kbs_token
                if self._auth._config.kbs_token is not None
                else []
            )
            kb_obj = retrieve(kbs, kb)

        if kb_obj is None:
            raise KeyError("Knowledge Box not found")
        self._auth._config.set_default_kb(kb_obj.id)
        self._auth._config.save()


class AsyncNucliaKBS:
    @property
    def _auth(self) -> AsyncNucliaAuth:
        auth = get_async_auth()
        return auth

    @accounts
    async def list(self, account: Optional[str] = None):
        if account is None:
            result = []
            accounts = (
                self._auth._config.accounts
                if self._auth._config.accounts is not None
                else []
            )
            for account_obj in accounts:
                if account_obj.slug is not None:
                    result.extend(await self._auth.kbs(account_obj.id, cached=False))
            self._auth._config.kbs = result
            self._auth._config.save()

            # List the Knowledge Boxes configured as Service Token
            result.extend(
                self._auth._config.kbs_token
                if self._auth._config.kbs_token is not None
                else []
            )
            return result
        else:
            matching_account = retrieve_account(
                self._auth._config.accounts or [], account
            )
            if not matching_account:
                raise ValueError("Account not found")
            return await self._auth.kbs(matching_account.id, cached=False)

    @accounts
    @account
    @zone
    async def add(
        self,
        slug: str,
        anonymization: Optional[str] = None,
        description: Optional[str] = None,
        learning_configuration: Optional[Dict[str, str]] = None,
        sentence_embedder: Optional[str] = None,
        title: Optional[str] = None,
        zone: Optional[str] = "europe-1",
        **kwargs,
    ):
        if not slug:
            raise ValueError("slug is required.")
        if not zone:
            raise ValueError("zone is required")
        path = get_regional_url(zone, KBS_ENDPOINT.format(account=kwargs["account_id"]))
        data = {
            "slug": slug,
            "anonymization": anonymization,
            "description": description,
            "learning_configuration": learning_configuration,
            "sentence_embedder": sentence_embedder,
            "title": title or slug,
            "zone": zone,
        }
        await self._auth._request("POST", path, data)
        return await self.get(slug, account_id=kwargs["account_id"], zone=zone)

    @accounts
    @account
    @zone
    async def get(
        self,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        **kwargs,
    ):
        zone = kwargs.get("zone")
        if not zone:
            raise ValueError("zone is required")
        if not id and not slug:
            raise ValueError("id or slug is required")
        if slug and not id:
            kbs = await self._auth.kbs(kwargs["account_id"], cached=False)
            kb_obj = retrieve(kbs, slug)
            if not kb_obj:
                raise ValueError("Knowledge Box not found")
            id = kb_obj.id
        path = get_regional_url(
            zone, KB_ENDPOINT.format(account=kwargs["account_id"], kb=id)
        )
        return await self._auth._request("GET", path)

    @accounts
    @account
    @zone
    async def delete(
        self,
        slug: Optional[str] = None,
        id: Optional[str] = None,
        **kwargs,
    ):
        zone = kwargs.get("zone")
        if not zone:
            raise ValueError("zone is required")
        if not id and not slug:
            raise ValueError("id or slug is required")
        if slug and not id:
            kbs = await self._auth.kbs(kwargs["account_id"], cached=False)
            kb_obj = retrieve(kbs, slug)
            if not kb_obj:
                raise ValueError("Knowledge Box not found")
            id = kb_obj.id
        path = get_regional_url(
            zone, KB_ENDPOINT.format(account=kwargs["account_id"], kb=id)
        )
        return await self._auth._request("DELETE", path)

    async def default(self, kb: str):
        kbs = self._auth._config.kbs if self._auth._config.kbs is not None else []
        kb_obj = retrieve(kbs, kb)
        if kb_obj is None:
            kbs = (
                self._auth._config.kbs_token
                if self._auth._config.kbs_token is not None
                else []
            )
            kb_obj = retrieve(kbs, kb)

        if kb_obj is None:
            raise KeyError("Knowledge Box not found")
        self._auth._config.set_default_kb(kb_obj.id)
        self._auth._config.save()



import pytest
from nucliadb_sdk.v2.exceptions import UnknownError

from nuclia.sdk.kb import AsyncNucliaKB, NucliaKB


def test_configuration(testing_config):
    kb = NucliaKB()
    kb.get_configuration()

    # Should raise 422 error because the generative_model "foobar" does not
    # exist, but it's ok as we are only testing the library here.
    with pytest.raises(UnknownError) as err:
        kb.update_configuration(
            generative_model="foobar",
            semantic_model="bar",
            visual_labeling="yes",
            anonymization_model="I do not exist either",
            ner_model="neither do I",
        )
    assert "422" in str(err.value)


@pytest.mark.asyncio
async def test_configuration_async(testing_config):
    kb = AsyncNucliaKB()
    await kb.get_configuration()

    # Should raise 422 error because the generative_model "foobar" does not
    # exist, but it's ok as we are only testing the library here.
    with pytest.raises(UnknownError) as err:
        await kb.update_configuration(
            generative_model="foobar",
            semantic_model="bar",
            visual_labeling="yes",
            anonymization_model="I do not exist either",
            ner_model="neither do I",
        )
    assert "422" in str(err.value)



from nucliadb_models.resource import Resource, ResourceList

from nuclia.decorators import agent
from nuclia.lib.agent import AgentClient, AsyncAgentClient


class NucliaAgentSessions:
    @agent
    def new(self, name: str, **kwargs) -> str:
        ac: AgentClient = kwargs["ac"]
        return ac.new_session(name)

    @agent
    def delete(self, session_uuid: str, **kwargs):
        ac: AgentClient = kwargs["ac"]
        return ac.delete_session(session_uuid)

    @agent
    def list(self, page: int = 0, page_size: int = 20, **kwargs) -> ResourceList:
        ac: AgentClient = kwargs["ac"]
        return ac.get_sessions(page=page, page_size=page_size)

    @agent
    def get(self, session_uuid: str, **kwargs) -> Resource:
        ac: AgentClient = kwargs["ac"]
        return ac.get_session(session_uuid)


class AsyncNucliaAgentSessions:
    @agent
    async def new(self, name: str, **kwargs) -> str:
        ac: AsyncAgentClient = kwargs["ac"]
        return await ac.new_session(name)

    @agent
    async def delete(self, session_uuid: str, **kwargs):
        ac: AsyncAgentClient = kwargs["ac"]
        return await ac.delete_session(session_uuid)

    @agent
    async def list(self, page: int = 0, page_size: int = 20, **kwargs) -> ResourceList:
        ac: AsyncAgentClient = kwargs["ac"]
        return await ac.get_sessions(page=page, page_size=page_size)

    @agent
    async def get(self, session_uuid: str, **kwargs) -> Resource:
        ac: AsyncAgentClient = kwargs["ac"]
        return await ac.get_session(session_uuid)
